"""The python file for offset portfolio calculation."""
import logging
import os
import json
import sys
import codecs
from types import StringType
import time
import platform
import hashlib
import locale
import threading

import natcap.opal
import natcap.invest
from natcap.invest.iui import executor as invest_executor
import shapely
import shapely.speedups
import shapely.wkb
import shapely.prepared
import shapely.geometry
import shapely.geos
from osgeo import ogr

LOGGER = logging.getLogger('natcap.opal.offsets')

# create a logging filter
class TimedLogFilter(logging.Filter):
    # only print a single log message if the time passed since last log message
    # >= the user-defined interval
    def __init__(self, interval):
        logging.Filter.__init__(self)
        self.interval = interval
        self._last_record_time = None

    def filter(self, record):
        current_time = time.time()
        if self._last_record_time is None:
            # if we've never printed a message since starting the filter, print
            # the message.
            self._last_record_time = current_time
            return True
        else:
            # Only log a message if more than <interval> seconds have passed
            # since the last record was logged.
            if current_time - self._last_record_time >= 5.0:
                self._last_record_time = current_time
                return True
            return False


def build_shapely_polygon(ogr_feature, prep=False, fix=False):
    geometry = ogr_feature.GetGeometryRef()
    try:
        polygon = shapely.wkb.loads(geometry.ExportToWkb())
    except shapely.geos.ReadingError:
        LOGGER.debug('Attempting to close geometry rings')
        # If the geometry does not form a closed circle, try to connect the
        # rings with the OGR function.
        geometry.CloseRings()
        polygon = shapely.wkb.loads(geometry.ExportToWkb())
        LOGGER.debug('Geometry fixed')

    if fix:
        polygon = polygon.buffer(0)
    if prep:
        polygon = shapely.prepared.prep(polygon)
    return polygon


def assert_files_exist(files):
    """Assert that all input files exist.

        files - a list of URIs to files on disk.

    This function raises IOError when a file is not found."""

    for file_uri in files:
        if not os.path.exists(file_uri):
            raise IOError('File not found: %s' % file_uri)


def _log_model(model_name, model_args, session_id=None):
    """Log information about a model run to a remote server.

    Parameters:
        model_name (string): a python string of the package version.
        model_args (dict): the traditional InVEST argument dictionary.

    Returns:
        None
    """

    logger = logging.getLogger('natcap.opal.utils._log_model')

    string = model_name[:]  # make a copy
    if 'palisades' in globals().keys() or 'palisades' in sys.modules:
        string += '.gui'
    else:
        string += '.cli'

    # if we're in a frozen environment, fetch the build information about the
    # distribution.
    if natcap.opal.is_frozen():
        # read build information from the local configuration file.
        # distribution_name
        # distribution_build_id
        #
        # Build a full model name and version string out of this info
        dist_data_file = os.path.join(natcap.opal.get_frozen_dir(),
            'dist_version.json')
        dist_data = json.load(open(dist_data_file))

        model_name = "%s.%s" % (dist_data['dist_name'], string)
        model_version = dist_data['build_id']
    else:
        # we're not in a built distribution, so we someone must be running this
        # from the command line.  In this case, we're obviously not in a
        # distribution, but we may be in the permitting repo, not just as a
        # standalone natcap.opal repo.
        model_name = string
        model_version = natcap.opal.__version__

    def _node_hash():
        """Returns a hash for the current computational node."""
        data = {
            'os': platform.platform(),
            'hostname': platform.node(),
            'userdir': os.path.expanduser('~')
        }
        md5 = hashlib.md5()
        # a json dump will handle non-ascii encodings
        md5.update(json.dumps(data))
        return md5.hexdigest()

    try:
        bounding_box_intersection, bounding_box_union = (
            invest_executor._calculate_args_bounding_box(model_args))

        payload = {
            'model_name': model_name,
            'invest_release': model_version,
            'node_hash': _node_hash(),
            'system_full_platform_string': platform.platform(),
            'system_preferred_encoding': locale.getdefaultlocale()[1],
            'system_default_language': locale.getdefaultlocale()[0],
            'bounding_box_intersection': str(bounding_box_intersection),
            'bounding_box_union': str(bounding_box_union),
            'session_id': session_id,
        }

        logging_server = invest_executor._get_logging_server()
        logging_server.log_invest_run(payload)
    except Exception as exception:
        # An exception was thrown, we don't care.
        logger.warn(
            'an exception encountered when logging %s', str(exception))

class VectorUnprojected(Exception):
    """An Exception in case a vector is unprojected"""
    pass

class DifferentProjections(Exception):
    """An exception in cast a set of dataets are not in the same projection."""
    pass

def assert_ogr_projections_equal(vector_uri_list):
    """Assert that all projections of the input OGR-compatible vectors are
    identical.

    Raises VectorUnprojected if a vector is inprojected.
    Raises DifferentProjections if projections differ."""

    vector_list = [(ogr.Open(v_uri), v_uri) for v_uri in vector_uri_list]
    vector_projections = []

    unprojected_vectors = set()

    for vector, vector_uri in vector_list:
        layer = vector.GetLayer()
        srs = layer.GetSpatialRef()
        if not srs.IsProjected():
            unprojected_vectors.add(vector_uri)
        vector_projections.append((srs, vector_uri))

    if len(unprojected_vectors) > 0:
        raise VectorUnprojected(
            "These vectors are unprojected: %s" % (unprojected_vectors))

    for index in range(len(vector_projections) - 1):
        if not vector_projections[index][0].IsSame(
                vector_projections[index + 1][0]):
            LOGGER.warn(
                "These two datasets might not be in the same projection."
                " The different projections are:\n\n'filename: %s'\n%s\n\n"
                "and:\n\n'filename:%s'\n%s\n\n",
                vector_projections[index][1],
                vector_projections[index][0].ExportToPrettyWkt(),
                vector_projections[index+1][1],
                vector_projections[index+1][0].ExportToPrettyWkt())

    for vector in vector_list:
        #Make sure the vector is closed and cleaned up
        vector = None
    vector_list = None
    return True

def write_csv(uri, fields, rows):
    """Write a utf-8 encoded CSV to URI.
        fields - a list of strings
        rows - a list of dictionaries, pointing fieldname to value"""

    def _cast_if_str(string):
        if type(string) is StringType:
            return unicode(string, 'utf-8')
        return string

    # write the rows to a CSV
    LOGGER.debug('Writing output CSV: %s', uri)
    csv_file_obj = codecs.open(uri, 'w', 'utf-8')

    # write the header
    sanitized_fieldnames = ['"%s"' % _cast_if_str(f) for f in fields]
    csv_file_obj.write("%s\n" % ','.join(sanitized_fieldnames))

    # write the rows
    for row in rows:
        row_fields = ['"%s"' % _cast_if_str(row[f]) for f in fields]
        csv_file_obj.write("%s\n" % ','.join(row_fields))
    csv_file_obj.close()

def sigfig(number, digits=3):
    """
    Round a number to the given number of significant digits.
    Example:
        >>> sigfig(1234, 3)
        1230

    number (int or float): The number to adjust
    digits=3 (int): The number of siginficant figures to retain

    Returns a number."""

    input_type = type(number)
    output_num = ''
    digit_count = 0
    period_found = False
    for digit in str(float(number)):
        if digit == '.':
            period_found = True
            output_num += '.'
        elif digit_count < digits:
            output_num += digit
            digit_count += 1
        else:
            output_num += '0'

    return input_type(float(output_num))





"""A module for the generation of report-related functions."""
import codecs
import csv
import json
import logging
import os
import tempfile
from types import FloatType
from types import IntType
from types import StringType
from types import UnicodeType

from osgeo import ogr
import shapely.geos
import shapely.validation

import adept_core
import natcap.opal.i18n
import offsets
import preprocessing
import utils

LOGGER = logging.getLogger('natcap.opal.reporting')
_ = natcap.opal.i18n.language.ugettext

def sigfig(number, places):
    """
    Round `number` to `places` significant digits.

    Parameters:
        number (int or float): A number to round.
        places (int): The number of places to round to.

    Returns:
        A number
    """

    # Passing a negative int to round() gives us a sigfig determination.
    # Example: round(12345, -2) = 12300
    ndigits = -int(len(str(abs(number)).split('.')[0]) - places)
    return round(number, ndigits)

def _recode(string):
    """Re-encode an input string as UTF-8 if it isn't already encoded.  If
    the input value is not a string, return the value as-is."""
    if type(string) is StringType:
        return string.decode('utf-8')
    elif type(string) is UnicodeType:
        return string.encode('utf-8')
    return string

def get_impact_data(municipalities_vector, impacts_vector, name_col, pop_col,
        service_mitrat, out_json_file=None):
    """Calculate per-impact information as it relates to municipalities.

        municipalities_vector - a URI to an OGR vector
        impacts_vector - a URI to an OGR vector
        name_col - a field name in the municipalities vector that contains the
            polygon name
        pop_col - a field name in the municipalities vector containing an
            integer representing the population of the polygon.
        service_mitrat - a python list of strings containing services to consider.
        out_json_file=None - a URI to where an output json file with this data
            should be written.
"""

    utils.assert_files_exist([municipalities_vector, impacts_vector])

    LOGGER.debug('Opening municipalities vector %s', municipalities_vector)
    muni_vector = ogr.Open(municipalities_vector)
    muni_layer = muni_vector.GetLayer()

    LOGGER.debug('Opening impacts vector: %s', impacts_vector)
    impact_index, impact_dict = preprocessing.build_spatial_index(impacts_vector)
    impacts_vector = ogr.Open(impacts_vector)
    impacts_layer = impacts_vector.GetLayer()

    services = service_mitrat.keys()
    LOGGER.debug('Considering services %s', services)

    # preprocess impacts so we don't have to retrieve information repeatedly
    # later on.
    impact_features = {}
    for impact_feature in impacts_layer:
        impact_fid = impact_feature.GetFID()
        service_dict = {}

        for service_name in services:
            impact = impact_feature.GetField(service_name)
            if impact is None:
                LOGGER.debug('Impact %s has %s impact of None', impact_fid,
                    service_name)
                impact = 0.0
            service_dict[service_name] = sigfig(impact, 3)
        impact_features[impact_fid] = service_dict
    impacts_layer.ResetReading()

    per_muni_data = {}

    for muni_feature in muni_layer:
        muni_fid = muni_feature.GetFID()
        muni_geom = muni_feature.GetGeometryRef()
        LOGGER.debug('Processing municipality %s', muni_fid)
        muni_name = muni_feature.GetField(name_col)
        muni_population = muni_feature.GetField(pop_col)

        try:
            muni_polygon = offsets.build_shapely_polygon(muni_feature)
        except shapely.geos.ReadingError:
            LOGGER.debug('Muni %s (%s) has invalid geometry: "%s". Skipping',
                muni_fid, muni_name, shapely.validation.explain_validity())
            continue


        muni_data = {
            'pop': muni_population,
        }

        impacts = {}
        template = '%s impact'
        for service in services:
            if service == 'nutrient':
                service = 'nitrogen'
            impacts[template % service.capitalize()] = 0.0

        num_impacts = 0
        skipped_impacts = []
        for impact_fid in impact_index.intersection(muni_polygon.bounds):
            if sum(impact_features[impact_fid].values()) == 0:
                skipped_impacts.append(impact_fid)
                continue

            impact_feature = impacts_layer.GetFeature(impact_fid)
            impact_geom = impact_feature.GetGeometryRef()

            # get the intersection geometry and area.
            # if the area is 0, we know that there's no intersection.
            intersection = impact_geom.Intersection(muni_geom)
            intersection_area = intersection.Area()

            if intersection_area > 0:
                num_impacts += 1
                # Find the proportion of the impact site that is in this
                # municipality.
                prop_area = intersection_area / impact_geom.Area()

                impact_data = impact_features[impact_fid]

                for service in services:
                    if service == 'nutrient':
                        label = 'Nitrogen'
                    else:
                        label = service.capitalize()
                    impact_amt = impact_data[service] * prop_area
                    mit_ratio = service_mitrat[service]
                    impacts[template % label] += impact_amt * mit_ratio

        if len(skipped_impacts) > 0:
            LOGGER.debug('%s impact sites had not impacts and were skipped: %s',
                len(skipped_impacts), skipped_impacts)
        impacts_layer.ResetReading()

        # If at least one of the service impacts is not 0.0 (the default), then
        # we want to initialize the impacts.
        if True in map(lambda x: x != 0.0, impacts.values()):
            muni_data['impacts'] = impacts
        per_muni_data[muni_name] = muni_data
    muni_layer.ResetReading()

    LOGGER.debug('Per-municipality impacts: %s', json.dumps(per_muni_data,
        sort_keys=True, indent=4))

    if out_json_file is not None:
        LOGGER.debug('Saving per-muni data to %s', out_json_file)
        json.dump(per_muni_data, open(out_json_file, 'w'), sort_keys=True,
            indent=4)

    return per_muni_data

def build_parcel_table(per_offset_data, total_impacts, out_csv,
        dist=None, include_aoi_column=True, include_subzone_column=True,
        suggested_parcels=[]):
    """Build up a list of dictionaries with the following attributes per row:
        row['Sediment'] - sediment load offset by parcel
        row['distance (km)'] - shortest distance between offset site and the
            closest impact site
        row['Area (ha)'] - the area of the offset site (in ha)
        row['Ecosystem type'] - a string containing the ecosystem name
        row['Nitrogen'] - nitrogen load offset by parcel
        row['Carbon'] - Carbon offset by the parcel
        row['% Impact (Sediment)'] - the % of total sediment impact offset by
            the parcel
        row['% Impact (Nitrogen)'] - the % of total nutrient impact offset by
            the parcel
        row['% Impact (Carbon)'] - the % of total carbon impact offset by
            the parcel
        row['parcel_id'] - the parcel ID of the offset parcel.
        row['Richness'] - the Richness score (if it's provided)
        row['Threat'] - the threat score (if it's provided)

        Parameters:
        per_offset_data - a python dictionary where the key is the parcel ID of
        the offset parcel in question, and the value is a dictionary with the
        following attributes:
            'Sediment' - the sediment load offset by the parcel
            'Distance' - the minimum distance (in m) from the offset site to
                the closest impact site.
            'Area' - the area of the offset parcel (in square m)
            'Ecosystem' - the full name of the ecosystem type of the offset
                parcel
            'Nitrogen' - the amount of Nitrogen load offset by the parcel
            'Carbon' - the amount of carbon offset by the parcel

        total_impacts - a python dictionary with the following values:
            'sediment' - a float, the total sediment impact
            'nutrient' - a float, the total nitrogen impact
            'carbon' - a float, the total carbon impact.
            'custom' - (optional) a float, the total custom ES impact.
        out_csv - the file to which the parcel table information should be
        saved.
        dist - a string, either natcap.opal.adept_core.DIST_OPAL or
            natcap.opal.adept_core.DIST_MAFE.  Changes the rendering of a
            cuople strings.
        include_aoi_column=True - Boolean.  Indicate whether to include the AOI
        column in the output tables.
        include_subzone_column=True (Boolean): Indicate whether to include the
        subzone column in the output table.
        suggested_parcels=[] (List of integers): integer feature indices indicate which
        parcels are recommended.  These parcel rows will have a 1 or 0 in them
        indicating whether they are required.

        Writes a dictionary to the URI at out_csv.
"""
    LOGGER.debug('Total impacts: %s', total_impacts)

    if dist is None:
        dist = adept_core.DIST_OPAL

    def percent(num, denom):
        """Calculate num/denom as a percentage between 0 and 100.
            num - some number.
            denom - some number.  If denom is 0, it will be converted to 1 to
                avoide a ZeroDivisionError.

            Returns a number."""
        if denom == 0:
            denom = 1.0

        # if the denominator (impact) is positive, no offsets are needed, so
        # the percentage returned should be 0.0
        if denom > 0:
            return 0.0
        elif num < 0.0 and denom < 0.0:
            # if BOTH the denominator (impact) AND the numerator (offset) are
            # negative, the returned percentage should be negative.
            num = float(num) * -1
            return num / denom * 100.0
        else:
            # If none of these special cases are met, return the absolute value
            # of the actual percentage.
            return abs(float(num) / denom) * 100

    fieldname_map = {
        'parcel_id': _('Potential offset patch ID code'),
        'ecosystem_type': _('Ecosystem type'),
        'sediment': _('Sediment retention offset'),
        'nitrogen': _('Nitrogen retention offset'),
        'carbon': _('Carbon storage offset'),
        '%_impact_sed': _('Percent of total impact offset (sediment)'),
        '%_impact_nut': _('Percent of total impact offset (nitrogen)'),
        '%_impact_car': _('Percent of total impact offset (carbon)'),
        'distance': _('Distance from impact site (km)'),
        'area': _('Total patch area (ha)'),
        'custom': _('Custom ecosystem service offset'),
        '%_impact_custom': _('Percent of total impact offset (custom ES)'),
        'lci': _('Landscape context index (LCI)'),
        'recommended': _('Suggested offset parcel'),
    }

    fieldnames = [
        fieldname_map['parcel_id'],
        fieldname_map['ecosystem_type'],
        fieldname_map['distance'],
        fieldname_map['area'],
    ]

    services = total_impacts.keys()[:]  # make a copy
    if 'custom' in services:
        services.remove('custom')

    num_services = 0
    for service_name in services:
        if service_name == 'nutrient':
            service_key = 'nitrogen'
        else:
            service_key = service_name
        field_name = fieldname_map[service_key]
        field_percent = fieldname_map['%s_impact_%s' % ('%',
            service_name[0:3].lower())]
        fieldnames.insert(2 + num_services, field_name)
        fieldnames.insert(3 + (2 * num_services), field_percent)
        num_services += 1

    # mapping parcel_data keys to column names.
    # If the parcel_data keys are present, they should be included in the
    # offset row.
    fields_to_check = {
        'City': _('Soft Boundary 3'),
        'Avoidance Areas': _('Avoidance Areas'),
        'Conservation Portfolio': _('Conservation Portfolio'),
        'Richness': _('Richness'),
        'Threat': _('Threat'),
        'Lci': fieldname_map['lci'],
    }

    if include_aoi_column:
        fields_to_check['Aoi'] = _('Soft Boundary 2')

    if include_subzone_column:
        fields_to_check['Subzone'] = _('Soft Boundary 1')

    # Assume we're running OPAL unless we specify otherwise.
    if dist == adept_core.DIST_MAFE:
        fields_to_check['Subzone'] = _('Hydrographic sub-zone')
        fields_to_check['Aoi'] = _('AOI (area of influence)')
        fields_to_check['City'] = _('Municipality')

    first_offset = per_offset_data.items()[0][1]  # 2nd item of 1st tuple in list
    LOGGER.debug('Checking for additional fields in  %s', first_offset)
    fields_to_copy = []
    for json_fieldname, csv_fieldname in fields_to_check.iteritems():
        if json_fieldname in first_offset:
            fieldnames.append(csv_fieldname)
            fields_to_copy.append((json_fieldname, csv_fieldname))

    # If we're also considering custom ecosystem services, add the correct
    # fields to the fieldnames list
    if 'Custom' in first_offset:
        fieldnames.insert(2 + len(services), fieldname_map['custom']) #5
        fieldnames.insert(3 + (2 * len(services)), fieldname_map['%_impact_custom']) #9

    LOGGER.debug('Also including fields (json, csv) %s', fields_to_copy)
    LOGGER.debug('Fieldnames: %s', fieldnames)

    suggested_parcels_set = set(suggested_parcels)
    rows = []
    for parcel_id, parcel_data in per_offset_data.iteritems():
        LOGGER.debug('parcel_data %s', parcel_data)

        data = {
            fieldname_map['distance']: sigfig(parcel_data['Distance'] / 1000.0, 3),
            fieldname_map['area']: parcel_data['Area'] / 10000.0,
            fieldname_map['ecosystem_type']: _recode(parcel_data['Ecosystem']),
            fieldname_map['parcel_id']: parcel_id,
            fieldname_map['recommended']: int(parcel_id in suggested_parcels_set),
        }
        for es_key in ['Sediment', 'Nitrogen', 'Carbon', 'Custom']:
            if es_key not in parcel_data:
                LOGGER.debug('key %s not in parcel_data', es_key)
                continue

            fieldname_key = es_key.lower()
            if es_key == 'Custom':
                shortened_name = es_key.lower()
            else:
                if es_key == 'Nitrogen':
                    shortened_name = 'nut'
                else:
                    shortened_name = es_key[0:3].lower()
            percent_key = '%s_impact_%s' % ('%', shortened_name)
            data[fieldname_map[fieldname_key]] = sigfig(parcel_data[es_key], 3)

            if fieldname_key == 'nitrogen':
                fieldname_key = 'nutrient'
            data[fieldname_map[percent_key]] = percent(parcel_data[es_key],
                total_impacts[fieldname_key])

        for key in [_('Threat'), _('Richness')]:
            try:
                data[key] = parcel_data[key]
            except KeyError:
                pass

        # copy over fields that we now know to exist.
        for json_field, csv_field in fields_to_copy:
            data[csv_field] = sigfig(parcel_data[json_field], 3)

        rows.append(data)

    utils.write_csv(out_csv, fieldnames, rows)

    columns = [
        {'name': fieldname_map['parcel_id'], 'total': False,
            'attr': {'class': 'parcel_id'}},
        {'name': fieldname_map['recommended'], 'total': False,
            'attr': {'class': 'tdcenter'}},
        {'name': fieldname_map['sediment'], 'total': True,
            'attr': {'class': 'round2'}} if 'sediment' in services else None,
        {'name': fieldname_map['nitrogen'], 'total': True,
            'attr': {'class': 'round2'}} if 'nutrient' in services else None,
        {'name': fieldname_map['carbon'], 'total': True,
            'attr': {'class': _('Carbon').lower() + ' round2'}} if 'carbon' in services else None,
        {'name': fieldname_map['%_impact_sed'], 'total': True,
            'attr': {'class': 'round2'}} if 'sediment' in services else None,
        {'name': fieldname_map['%_impact_nut'], 'total': True,
            'attr': {'class': 'round2'}} if 'nutrient' in services else None,
        {'name': fieldname_map['%_impact_car'], 'total': True,
            'attr': {'class': 'round2'}} if 'carbon' in services else None,
        {'name': fieldname_map['area'], 'total': True,
            'attr': {'class': 'area round2'}},
        {'name': fieldname_map['ecosystem_type'], 'total': False,
            'attr': {'class': 'eco_type'}},
        {'name': fieldname_map['distance'], 'total': False,
            'attr': {'class': 'round2'}},
    ]
    new_columns = []
    for col in columns:
        if col is None:
            continue
        new_columns.append(col)
    columns = new_columns

    if 'Custom' in first_offset:
        columns.insert(4, {'name': fieldname_map['custom'],
            'total': True, 'attr': {'class': _('Custom').lower() + ' round2'}})
        columns.insert(8, {'name': fieldname_map['%_impact_custom'],
            'total': True, 'attr': {'class': 'round2'}})

    # some fields should be ints.  If the current field isn't one of these
    # fields, it should be rounded to 2 digits.
    fields = [fields_to_check['Subzone'], fields_to_check['City']]
    if 'Aoi' in fields_to_check:
        fields.insert(0, fields_to_check['Aoi'])

    for json_field, csv_field in fields_to_copy:
        if csv_field in fields:
            round_digit = '0'
            center_class = 'tdcenter'
        else:
            round_digit = '2'
            center_class = ''

        column_data = {
            'name': csv_field,
            'total': False,
            'attr': {'class': 'round%s %s' % (round_digit, center_class)},
        }
        columns.append(column_data)

    reporting_config = {
        'type': 'table',
        'section': 'body',
        'sortable': True,
        'checkbox': True,
        'totals': True,
        'attributes': {'id': 'parcel_table'},
        'data_type': 'dictionary',
        'columns': columns,
        'key': fieldname_map['parcel_id'],
        'data': rows,
    }
    return reporting_config

def impacted_parcels_table(impact_sites, natural_parcels, csv_uri):
    """Determine which natural parcels have been impacted and build a
    reporting table based on the results.

        impact_sites - a URI to an OGR vector with impact sites
        natural_parcels - a URI to an OGR vector with natural parcels.
            Must have vegetation type and LCI, but may also have Threat and
            Richness columns.  If threat and richness are provided, they will
            be included in the resulting table.
        csv_uri - a URI to an output CSV.

    Returns a reporting args dictionary with appropriate data."""

    # create a temp folder
    # build column names
    # locate intersecting natural polygons, saving them to the temp dir.
    # for natural polygon in intersecting polygons,
    #    build up a row with relevant data from the current natural polygon
    # remove the temp folder.
    # Write the row/column data to a CSV
    # return an appropriate dictionary for reporting.

    utils.assert_files_exist([impact_sites])

    temp_dir = tempfile.mkdtemp()

    impacted_natural_parcels = os.path.join(temp_dir, 'impact_parcels.shp')
    preprocessing.locate_intersecting_polygons(natural_parcels, impact_sites,
        impacted_natural_parcels)

    parcels_vector = ogr.Open(impacted_natural_parcels)
    parcels_layer = parcels_vector.GetLayer()

    layer_fields = map(lambda f: f.GetName(), parcels_layer.schema)
    LOGGER.debug('Found fields %s', layer_fields)

    field_functions = {
        _('Impacted patch code'): lambda p: p.GetFID(),
        _('Ecosystem type'): lambda p: p.GetField('ecosystem'),
        _('Mitigation ratio'): lambda p: p.GetField('mit_ratio'),
    }

    if 'LCI' in layer_fields:
        field_functions[_('Landscape Context Index (LCI)')] = lambda p: round(
            p.GetField('LCI'), 2)

    if 'Richness' in layer_fields:
        field_functions['Richness score'] = lambda p: round(
            p.GetField('Richness'), 2)

    if 'Threat' in layer_fields:
        field_functions['Threat score'] = lambda p: round(p.GetField('Threat'),
            2)

    LOGGER.debug('Buliding up shapely impact geometries')
    # build up a list of prepared shapely geometries for fast intersections.
    impacts_vector = ogr.Open(impact_sites)
    impacts_layer = impacts_vector.GetLayer()
    shapely_impacts = [offsets.build_shapely_polygon(f) for f in impacts_layer]

    rows = []
    for natural_parcel in parcels_layer:
        # get the columns we know exist, based on our earlier tests
        row_data = {}
        for label, value_func in field_functions.iteritems():
            _data_value = value_func(natural_parcel)
            row_data[label] = _recode(_data_value)

        # now, add in the impacted area.  This is the sum of (area of the
        # intersection of each impact geometry with the natural parcel).
        offset_polygon = offsets.build_shapely_polygon(natural_parcel)
        prep_offset_polygon = shapely.prepared.prep(offset_polygon)
        intersection_area = 0.0
        for impact_polygon in shapely_impacts:
            if prep_offset_polygon.intersects(impact_polygon):
                intersection_area += offset_polygon.intersection(impact_polygon).area

        # convert from m^2 to ha.
        intersection_area /= 10000.0

        row_data[_('Patch area impacted (ha)')] = sigfig(round(intersection_area, 2), 3)
        row_data[_('Required offset area (ha)')] = sigfig(round(intersection_area *
            natural_parcel.GetField('mit_ratio'), 2), 3)

        rows.append(row_data)

    # order the fieldnames
    fieldnames = [
        _('Impacted patch code'),
        _('Patch area impacted (ha)'),
        _('Mitigation ratio'),
        _('Required offset area (ha)'),
        _('Ecosystem type'),
    ]
    if 'LCI' in layer_fields:
        fieldnames.insert(1, _('Landscape Context Index (LCI)'))

    for key in field_functions.keys():
        if key not in fieldnames:
            fieldnames.insert(-1, key)

    utils.write_csv(csv_uri, fieldnames, rows)

    # start off the html columns by getting the names
    html_columns = [{'name': c, 'total': False} for c in fieldnames]

    reporting_dict = {
        'type': 'table',
        'section': 'body',
        'sortable': True,
        'checkbox': False,
        'total': False,
        'data_type': 'dictionary',
        'key': 'FID',
        'columns': html_columns,
        'data': rows,
    }
    return reporting_dict

def es_benefits_table(services):
    """Generate a reporting dictionary for the Ecosystem Services Benefits
    table (formerly called the "municipaities table").

        services - a list of services, a combination of 'sediment',
            'nutrient', and/or 'custom' (case-sensitive)

    returns a reporting table configuration dictionary."""

    do_nutrient = 'nutrient' in services
    do_sediment = 'sediment' in services
    do_custom = 'custom' in services

    columns = [
        {
            'name': _('Population center'),
            'attr': {'class': 'services_pop_name'},
            'total':False
        },
        {
            'name': _('Population size'),
            'attr': {'class': 'services_pop_count'},
            'total': False
        },
        {
            'name': _('Sediment impact'),
            'total': False,
            'attr': {'class':'services_impact_sediment impacts round2'}
        } if do_sediment else None,
        {
            'name': _('Nitrogen impact'),
            'total': False,
            'attr':{'class':'services_impact_nitrogen impacts round2'}
        } if do_nutrient else None,
        {
            'name': _('Custom impact'),
            'total': True,
            'attr': {'class': 'services_impact_custom impacts round2'}
        } if do_custom else None,
        {
            'name': _('Sediment offset'),
            'total': False,
            'attr':{'class':'services_offset_sediment offsets round2'}
        } if do_sediment else None,
        {
            'name': _('Nitrogen offset'),
            'total': False,
            'attr':{'class':'services_offset_nitrogen offsets round2'}
        } if do_nutrient else None,
        {
            'name': _('Custom offset'),
            'total': True,
            'attr': {'class': 'services_offset_custom offsets round2'}
        } if do_custom else None,
        {
            'name': _('Net sediment retention benefits'),
            'total': False,
            'attr':{'class':'services_net_sediment net round2'}
        } if do_sediment else None,
        {
            'name': _('Net nitrogen retention benefits'),
            'total': False,
            'attr':{'class':'services_net_nitrogen net round2'}
        } if do_nutrient else None,
        {
            'name': _('Net custom ES benefits'),
            'total': True,
            'attr': {'class': 'services_net_custom net round2'}
        } if do_custom else None,
        {
            'name': _('Net sediment x population'),
            'total': False,
            'attr': {'class': 'services_netadj_sediment population round2'}
        } if do_sediment else None,
        {
            'name': _('Net nitrogen x population'),
            'total': False,
            'attr': {'class': 'services_netadj_nitrogen population round2'}
        } if do_nutrient else None,
        {
            'name': _('Net custom ES x population'),
            'total': False,
            'attr': {'class': 'services_netadj_custom population round2'}
        } if do_custom else None
    ]

    # build a list of columns based on whether those columns should be included
    # Column value is None if we should not.
    included_columns = []
    for possible_column in columns:
        if possible_column is not None:
            included_columns.append(possible_column)

    report_dict = {
        'type': 'table',
        'section': 'body',
        'sortable': False,
        'checkbox': False,
        'total': False,
        'attributes': {'id': 'muni_table'},
        'data_type': 'dictionary',
        'columns': included_columns,
        'data': [],
        'key': 'pop_group',
    }

    return report_dict

def global_benefits_table(add_custom, total_impacts):
    """Build the global benefits table, adding in custom ES columns if
    indicated.

    add_custom - a boolean indicating whether to add custom ES columns.
    total_impacts - a python dictionary containing at least these entries:
        'carbon' - (required) a float.
        'custom' - (required if add_custom is True) a float.

    Returns an invest_natcap.reporting compatible table dictionary.
    """

    benefits_table = {
        'type': 'table',
        'section': 'body',
        'sortable': True,
        'total': False,
        'attributes': {'id': 'global_benefits_table'},
        'data_type': 'dictionary',
        'key': _('Carbon impact'),
        'columns': [
            {
                'name': _('Service type'),
                'total': False,
                'attr': {'class': 'service_name'}
            },
            {
                'name': _('Impacts to service'),
                'total': True,
                'attr': {'class': 'impact round2'}
            },
            {
                'name': _('Selected offset'),
                'total': False,
                'attr': {'class': 'required_offset round2'}
            },
            {
                'name': _('Net service'),
                'total': False,
                'attr': {'class': 'net round2'}
            },
        ],
        'data': []
    }
    if 'carbon' in total_impacts:
        benefits_table['data'].append(
            {
                _('Service type'): _('Carbon'),
                _('Impacts to service'): sigfig(total_impacts['carbon'], 3),
                _('Selected offset'): 0.0,
                _('Net service'): sigfig(total_impacts['carbon'], 3),
            }
        )

    if add_custom is True:
        benefits_table['data'].append({
            _('Service type'): _('Custom'),
            _('Impacts to service'): sigfig(total_impacts['custom'], 3),
            _('Selected offset'): 0.0,
            _('Net service'): sigfig(0.0 - total_impacts['custom'], 3)
        })

    return benefits_table

def write_per_offset_csv(per_offset_data, out_csv):
    """Create a CSV documenting the benefits of each hydrological service
    granted to each serviceshed.

    per_offset_data - a dictionary with integer keys (representing a parcel ID)
        mapping to dictionaries of values containing at least the following
        entries:
            "Nitrogen" - a float (the nitrogen offset of this parcel)
            "Sediment" - a float (the sediment offset of this parcel)
            "municipalities" - a dictionary, with string keys indicating the
                names of affected servicesheds and values being a float between
                0.0 and 1.0.
    out_csv - a URI to where the csv should be saved."""

    if len(per_offset_data) == 0:
        LOGGER.debug('No offset parcels to calculate with.  Returning')
        return


    csv_file = codecs.open(out_csv, 'w', 'utf-8')
    def _write(row_data):
        row_string = ','.join(row_data)
        csv_file.write(row_string + '\n')

    fieldnames = ['Offset parcel ID', 'Serviceshed name']

    # check which services are included so we known which columns to use
    known_services = []
    sample_offset = per_offset_data.values()[0]
    for fieldname in ['Sediment', 'Nitrogen', 'Custom']:
        if fieldname in sample_offset:
            fieldnames.append(fieldname)
            known_services.append(fieldname)
    _write(fieldnames)

    for parcel_key, parcel_data in per_offset_data.iteritems():
        servicesheds = parcel_data['municipalities']
        for serviceshed_name, percent_overlap in servicesheds.iteritems():

            row_data = [parcel_key, serviceshed_name]
            for service_name in known_services:
                service_amt = percent_overlap * parcel_data[service_name]
                row_data.append(sigfig(service_amt, 3))

            def _smart_cast(string):
                """Cast to UTF-8 if the string is not already UTF-8"""
                if type(string) in [IntType, FloatType]:
                    return str(string)

                if type(string) is not UnicodeType:
                    return unicode(string, 'utf-8')
                return string

            _write(map(_smart_cast, row_data))

    csv_file.close()

def es_impacts_table(service_impacts, service_mitrat):
    """Create the ES impacts table for the report.

    service_impacts - a python dict mapping service names (any of 'carbon',
        'nutrient', 'sediment', or 'custom') to numeric service impact
        quantities.
    service_mitrat - a python dict mapping service names (any of those
        mentioned for the service_impacts input) to numeric service mitigation
        ratios.

    Returns a reporting-compatible dictionary representing the table"""

    table_strings = {
        'service_name': _('Service name'),
        'service_impact': _('Impact to service'),
        'mit_ratio': _('Mitigation ratio'),
        'required_offset': _('Required offset value'),
        'carbon': _('Carbon'),
        'nutrient': _('Nutrient'),
        'sediment': _('Sediment'),
        'custom': _('Custom ecosystem service'),
    }

    table_columns = [
        {
            'name': table_strings['service_name'],
            'total': False,
        },
        {
            'name': table_strings['service_impact'],
            'total': False,
            'attr': {'class': 'round2'},
        },
        {
            'name': table_strings['mit_ratio'],
            'total': False,
            'attr': {'class': 'round2'},
        },
        {
            'name': table_strings['required_offset'],
            'total': False,
            'attr': {'class': 'round2'},
        },
    ]

    def _calc_req_offset(service_impact, mit_ratio):
        if service_impact > 0:
            return 0.0
        return abs(service_impact * mit_ratio)

    table_rows = []
    for service_name, service_impact in service_impacts.iteritems():
        mit_ratio = service_mitrat[service_name]
        required_offset = _calc_req_offset(service_impact, mit_ratio)
        row_data = {
            table_strings['service_name']: table_strings[service_name],
            table_strings['service_impact']: sigfig(service_impact, 3),
            table_strings['mit_ratio']: mit_ratio,
            table_strings['required_offset']: sigfig(required_offset, 3),
        }
        table_rows.append(row_data)

    # sort the table rows based on the translated service name
    table_rows = sorted(table_rows,
        key=lambda x: x[table_strings['service_name']])

    reporting_dict = {
        'type': 'table',
        'section': 'body',
        'sortable': True,
        'checkbox': False,
        'total': False,
        'data_type': 'dictionary',
        'columns': table_columns,
        'data': table_rows,
        'position': 1,
    }
    return reporting_dict




import unittest
import time
import os
import sys

import natcap.opal
from natcap.opal import utils

class UtilsTest(unittest.TestCase):
    def test_sigfig(self):
        self.assertEqual(utils.sigfig(123456.1234, 3), 123000.0)
        self.assertEqual(utils.sigfig(1.12345, 3), 1.12)
        self.assertEqual(utils.sigfig(1234, 3), 1230)


class InitTest(unittest.TestCase):
    def test_local_dir_frozen(self):
        sys.frozen = True

        # simulate executable in src directory
        sys.executable = os.path.abspath(os.path.join(
            os.path.dirname(natcap.opal.__file__), '..', '..'))
        filepath = os.path.join(os.path.dirname(natcap.opal.__file__),
                                'report_data', 'table_style.css')
        self.assertEqual(natcap.opal.local_dir(filepath),
                         '/var/natcap/opal/dir')




import unittest
import glob
import tempfile
import os
import shutil
import inspect

from shapely.geometry import Polygon
import invest_natcap.testing
import numpy

import natcap.opal.i18n
from natcap.opal.tests import vector, raster, COLOMBIA_SRS, COLOMBIA_GEOTRANSFORM
from natcap.opal import adept_core

def square(center_coords, side_len):
    x, y = center_coords
    half_len = side_len / 2.0
    bl = (x - half_len, y - half_len)
    br = (x + half_len, y - half_len)
    tr = (x + half_len, y + half_len)
    tl = (x - half_len, y + half_len)

    return Polygon([bl, br, tr, tl, bl])

def subdivide(polygon, square_side_len):
    # polygon - a shapely polygon
    subzones = []
    polygon_x_min = int(min(map(lambda x: x[0], polygon.exterior.coords)))
    polygon_x_max = int(max(map(lambda x: x[0], polygon.exterior.coords)))
    polygon_y_min = int(min(map(lambda x: x[1], polygon.exterior.coords)))
    polygon_y_max = int(max(map(lambda x: x[1], polygon.exterior.coords)))

    for x in range(polygon_x_min, polygon_x_max, square_side_len):
        for y in range(polygon_y_min, polygon_y_max, square_side_len):
            x_center = x + square_side_len / 2
            y_center = y + square_side_len / 2
            subzone = square((x_center, y_center), square_side_len)
            subzones.append(subzone)

    return subzones

def named_workspace(clean=True, index=1):
    """Create a folder that is the same name as the function that called this
    function.  The new directory is created in the CWD.

        clean=True - a boolean.  If True, the workspace folder, if it already
            exists on disk, will be recursively removed before returning.
        index=1 - an integer, representing the call stack depth to query for
            the function name for the workspace. 0 would be named_workspace, 1
            would be the function that called named_workspace.

    Returns a string URI to the new, named folder.
"""
    dir_name = os.path.join(os.getcwd(), inspect.stack()[index][3])
    if clean and os.path.exists(dir_name):
        shutil.rmtree(dir_name)
    return dir_name

class SmokeTest(invest_natcap.testing.GISTest):
    def setUp(self):
        natcap.opal.i18n.language.set('en')
        # make directories on disk in which to store internal data
        self._data_base_dir = tempfile.mkdtemp()
        internal_static_data = os.path.join(self._data_base_dir, 'data',
            'colombia_static_data')
        internal_tool_data = os.path.join(self._data_base_dir, 'data',
            'colombia_tool_data')
        for dirname in [internal_static_data, internal_tool_data]:
            os.makedirs(dirname)

        # servicesheds
        servicesheds = []
        for x in range(0, 20000, 10000):
            for y in range(0, 40000, 10000):
                bl = (x, y)
                br = (x + 10000, y)
                ur = (x + 10000, y + 10000)
                ul = (x, y + 10000)
                servicesheds.append(Polygon([bl, br, ur, ul, bl]))
        fields = {
            'pop_center': str,
            'pop_size': int,
        }
        field_values = [{'pop_center': str(i), 'pop_size': i * 1000} for i in range(len(servicesheds))]
        servicesheds_filename = os.path.join(internal_tool_data,
            'ssheds_Col_new.shp')
        self.servicesheds = vector(servicesheds, COLOMBIA_SRS, fields,
            field_values, format='ESRI Shapefile', filename=servicesheds_filename)

        # municipalities
        municipalities = []
        for x in range(0, 20000, 5000):
            bl = (x, 0)
            br = (x + 5000, 0)
            ur = (x + 5000, 40000)
            ul = (x, 40000)
            municipalities.append(Polygon([bl, br, ur, ul, bl]))
        municipalities_filename = os.path.join(internal_tool_data,
            'Municipalities.shp')
        self.municipalities = vector(municipalities, COLOMBIA_SRS,
            format='ESRI Shapefile',
            filename=municipalities_filename)

        # make max search area (hydrozone) polygon
        max_search_area = Polygon([(0, 0), (0, 40000), (20000, 40000), (20000, 0), (0, 0)])
        fields = {'zone': str}
        field_values = [{'zone': 'Hydrozone_A'}]
        self.search_vector = vector([max_search_area], COLOMBIA_SRS,
            fields=fields, features=field_values, format='ESRI Shapefile')

        # make a series of polygons to make up hydrosubzones
        subzones = subdivide(max_search_area, 5000)
        field_values = field_values*len(subzones)
        self.hydro_subzones = vector(subzones, COLOMBIA_SRS, fields,
            field_values, format='ESRI Shapefile')

        # make a sample impacts set of polygons
        impact_data = [
            ((12500, 32500), 2000),
            ((14500, 34500), 2000),
            ((12500, 27500), 1000),
            ((13000, 29500), 1000),
            ((10000, 26500), 1500),
            ((9000, 28000), 1000),
        ]
        impacts = []
        for impact_center, side_len in impact_data:
            impact_site = square(impact_center, side_len)
            impacts.append(impact_site)
        fields = {
            'FID': int,
        }
        field_values = [{'FID': i} for i in range(len(impact_data))]
        self.impacts = vector(impacts, COLOMBIA_SRS, fields, field_values,
            format='ESRI Shapefile')

        # make a natural ecosystems vector.
        # this need not be especially complex, but it should probably go outside the
        # bounds of the hydrozone to exercise the LCI parcel aggregation functionality
        # and should definitely intersect at least one of the impact sites.
        ecosystems = []
        new_width = 2000
        half_width = new_width / 2
        for subzone in subzones:
            polygon_x_min = int(min(map(lambda x: x[0], subzone.exterior.coords)))
            polygon_x_max = int(max(map(lambda x: x[0], subzone.exterior.coords)))
            polygon_y_min = int(min(map(lambda x: x[1], subzone.exterior.coords)))
            polygon_y_max = int(max(map(lambda x: x[1], subzone.exterior.coords)))

            x_med = (polygon_x_max - polygon_x_min) / 2 + polygon_x_min
            y_med = (polygon_y_max - polygon_y_min) / 2 + polygon_y_min

            new_coords = []
            for x_coord, y_coord in subzone.exterior.coords:
                if x_coord < x_med:
                    out_x_coord = x_med - half_width
                else:
                    out_x_coord = x_med + half_width

                if y_coord < y_med:
                    out_y_coord = y_med - half_width
                else:
                    out_y_coord = y_med + half_width

                new_coords.append((out_x_coord, out_y_coord))

            ecosys_polygon = Polygon(new_coords)
            ecosystems.append(ecosys_polygon)

        ecosystem_fields = {
            'ecosystem': str,
            'mit_ratio': float
        }
        ecosystem_field_values = [{'ecosystem': 'Eco_A',
            'mit_ratio': 4.5}] * len(ecosystems)
        self.ecosystems = vector(ecosystems, COLOMBIA_SRS, ecosystem_fields,
            ecosystem_field_values, format='ESRI Shapefile')

        # make a user-defined example AOI.
        area_of_influence = Polygon([(10000, 25000), (17500, 27500),
            (17500, 36000), (10000, 36000), (6000, 27250), (10000, 25000)])
        self.aoi = vector([area_of_influence], COLOMBIA_SRS, format='ESRI Shapefile')

        # avoidance areas
        avoid_a = Polygon([(14500, 35000), (17500, 36000), (15000, 37000),
            (14500, 35000)])
        self.avoidance_areas = vector([avoid_a], COLOMBIA_SRS, format='ESRI Shapefile')

        # conservation portfolio
        conserve_a = Polygon([(5000, 20000), (15000, 20000), (10000, 26500),
            (5000, 20000)])
        self.conservation_portfolio = vector([conserve_a], COLOMBIA_SRS,
            format='ESRI Shapefile')

        # previously granted impact
        prev_impact = square((13000, 29500), 750)
        self.previous_impact = vector([prev_impact], COLOMBIA_SRS,
            format='ESRI Shapefile')

        # Threat data
        # hydrozone is 20000 pixels wide, 40000 pixels tall
        # Filling with a constant value is OK ... the offset selection will
        # accept parcels with a threat score of at least the same as the
        # impacted parcel.
        zero_matrix = numpy.zeros((400, 200))
        zero_matrix.fill(1)
        self.threat = raster(zero_matrix, COLOMBIA_SRS,
            COLOMBIA_GEOTRANSFORM(100, 100, (0, 0)), -1)

        # Richness data
        # hydrozone is 20000 pixels wide, 40000 pixels tall
        # Filling with a constant value is OK ... the offset selection will
        # accept parcels with a threat score of at least the same as the
        # impacted parcel.
        self.richness = raster(zero_matrix, COLOMBIA_SRS,
            COLOMBIA_GEOTRANSFORM(100, 100, (0, 0)), -1)

        # Sediment static data.
        ones_matrix = numpy.zeros((400, 200))
        ones_matrix.fill(1)

        # make directories on disk for storing 'user-defined' static data.
        self.sediment_static_data = tempfile.mkdtemp()
        self.nutrient_static_data = tempfile.mkdtemp()
        self.carbon_static_data = tempfile.mkdtemp()
        self.custom_static_data = tempfile.mkdtemp()

        static_datas = [
            ('sediment', self.sediment_static_data, True),
            ('nutrient', self.nutrient_static_data, True),
            ('carbon', self.carbon_static_data, False),
            ('custom', self.custom_static_data, True),
        ]
        for model_name, data_dir, routed in static_datas:
            for scenario in ['paved', 'bare', 'protection']:
                sc_filename = lambda x: "%s_%s_static_map%s.tif" % (model_name,
                    scenario, x)

                user_uri = os.path.join(data_dir, sc_filename(''))
                internal_uri = os.path.join(internal_static_data,
                    sc_filename('_lzw'))
                for uri in [user_uri, internal_uri]:
                    raster(ones_matrix, COLOMBIA_SRS, COLOMBIA_GEOTRANSFORM(100,
                        100, (0, 0)), -1, filename=uri)

                # if this is a routed model, make a percent-to-stream raster.
                # But opnly do this if we're running nutrient.
                if routed and model_name in ['nutrient']:
                    pts_filename = "%s_%s_pts.tif" % (model_name, scenario)
                    user_uri = os.path.join(data_dir, pts_filename)
                    internal_uri = os.path.join(internal_static_data,
                        pts_filename)
                    for uri in [user_uri, internal_uri]:
                        raster(ones_matrix, COLOMBIA_SRS,
                            COLOMBIA_GEOTRANSFORM(100, 100, (0, 0)), -1,
                            filename=uri)

        self.workspace = tempfile.mkdtemp()

        self.args = {
            'workspace_dir': self.workspace,
            'project_footprint_uri': self.impacts,
            'impact_type': 'Road/Paved',
            'ecosystems_map_uri': self.ecosystems,
            'search_areas_uri': self.hydro_subzones,
            'data_dir': self._data_base_dir,
        }

    def tearDown(self):
        md5sum_logfile = self.__class__.__name__ + '-' + self._testMethodName + '.log'
        natcap.opal.tests.record_workspace_md5sums(self.workspace, md5sum_logfile)

        folders_to_rm = [
            self.sediment_static_data,
            self.nutrient_static_data,
            self.carbon_static_data,
            self.custom_static_data,
            self.workspace,
            self._data_base_dir,
        ]
        for folder_uri in folders_to_rm:
            try:
                shutil.rmtree(folder_uri)
            except OSError:
                pass

    def test_smoke(self):
        self.args['workspace_dir'] = 'smoke_test_workspace'
        adept_core.execute(self.args)

    def test_smoke_es(self):
        natcap.opal.i18n.language.set('es')
        self.test_smoke()
        if os.path.exists('es_workspace'):
            shutil.rmtree('es_workspace')
        shutil.copytree(self.workspace, 'es_workspace')
        print natcap.opal.i18n.language.available_langs
        print natcap.opal.i18n.LOCALE_DIR

    def test_smoke_custom_es(self):
        self.args['custom_static_maps'] = self.custom_static_data
        self.assertRaises(KeyError, adept_core.execute, self.args)

        self.args['custom_servicesheds'] = 'unknown'
        self.assertRaises(AssertionError, adept_core.execute, self.args)

        self.args['custom_servicesheds'] = 'global'
        adept_core.execute(self.args)

    def test_smoke_user_es(self):
        self.args['carbon_static_maps'] = self.carbon_static_data
        self.args['sediment_static_maps'] = self.sediment_static_data
        self.args['nutrient_static_maps'] = self.nutrient_static_data

        adept_core.execute(self.args)

    def test_smoke_wrong_impact(self):
        self.args['impact_type'] = 'something'
        self.assertRaises(AssertionError, adept_core.execute, self.args)

    def test_smoke_no_services(self):
        self.args['do_carbon'] = False
        self.args['do_sediment'] = False
        self.args['do_nutrient'] = False
        self.assertRaises(RuntimeError, adept_core.execute, self.args)

    def test_smoke_sed_only(self):
        self.args['do_carbon'] = False
        self.args['do_sediment'] = True
        self.args['do_nutrient'] = False
        adept_core.execute(self.args)

    def test_smoke_carbon_only(self):
        self.args['do_carbon'] = True
        self.args['do_sediment'] = False
        self.args['do_nutrient'] = False
        adept_core.execute(self.args)

    def test_smoke_carbon_custom(self):
        self.args['do_carbon'] = True
        self.args['do_sediment'] = False
        self.args['do_nutrient'] = False
        self.args['custom_static_maps'] = self.custom_static_data
        self.args['custom_servicesheds'] = 'global'
        adept_core.execute(self.args)

    def test_smoke_sediment_custom(self):
        self.args['do_carbon'] = False
        self.args['do_sediment'] = True
        self.args['do_nutrient'] = False
        self.args['custom_static_maps'] = self.custom_static_data
        self.args['custom_servicesheds'] = 'hydrological'
        adept_core.execute(self.args)

    def test_smoke_sediment_custom_global(self):
        self.args['do_carbon'] = False
        self.args['do_sediment'] = True
        self.args['do_nutrient'] = False
        self.args['custom_static_maps'] = self.custom_static_data
        self.args['custom_servicesheds'] = 'global'
        adept_core.execute(self.args)

    def test_smoke_threat_richness(self):
        self.args['threat_map'] = self.threat
        self.args['species_richness_map'] = self.threat
        self.args['do_carbon'] = True
        self.args['do_sediment'] = True
        self.args['do_nutrient'] = True
        adept_core.execute(self.args)

    def test_user_municipalities(self):
        # assert failure when bad municpalities provided.
        self.args['municipalities_uri'] = 'does_not_exist_ljhkslfg'
        self.assertRaises(RuntimeError, adept_core.execute, self.args)

        # remove the workspace.  Files have been written to this URI, so we
        # need to clean this up before running this again.
        try:
            shutil.rmtree(self.workspace)
        except OSError:
            pass

        # assert failure when neither the internal municipalities nor the
        # user-defined municipalities exist on disk.
        # This means moving the Municipalities vector to a known location so
        # that it's not in the internal tool data dir.
        temp_dir = tempfile.mkdtemp()
        moved_files = []
        municipalities_base = os.path.splitext(self.municipalities)[0]
        for src_filename in glob.glob('%s.*' % municipalities_base):
            dest_filename = os.path.join(temp_dir,
                os.path.basename(src_filename))
            shutil.move(src_filename, dest_filename)
            moved_files.append((src_filename, dest_filename))

        # remove the user-provided municipalities args entry to force the
        # fetching of internal municipalities
        del self.args['municipalities_uri']
        self.assertRaises(RuntimeError, adept_core.execute, self.args)

        # remove the workspace.  Files have been written to this URI, so we
        # need to clean this up before running this again.
        try:
            shutil.rmtree(self.workspace)
        except OSError:
            pass

        # now, move the files back to their correct location and provide the
        # correct municipalities URI.
        for orig_uri, moved_uri in moved_files:
            shutil.move(moved_uri, orig_uri)

        self.args['municipalities_uri'] = self.municipalities
        adept_core.execute(self.args)

    def test_invalid_impacts(self):
        invalid_impacts_vector = vector([
            Polygon([(-100, 0), (-100, -100), (-200, -100), (-200, 0),
                (-100, 0)])],
            COLOMBIA_SRS, format='ESRI Shapefile')

        self.args['project_footprint_uri'] = invalid_impacts_vector
        self.assertRaises(adept_core.InvalidImpactsVector, adept_core.execute, self.args)

    def test_impacts_without_biodiversity(self):
        impacts_without_natural_parcels = vector([
            Polygon([(4000, 4000), (6000, 4000), (6000, 6000), (4000, 6000),
                (4000, 4000)])], COLOMBIA_SRS, fields={'FID': int},
            features=[{'FID': 1}], format='ESRI Shapefile')
        self.args['project_footprint_uri'] = impacts_without_natural_parcels
        adept_core.execute(self.args)

class SmokeTestOPAL(SmokeTest):
    def setUp(self):
        SmokeTest.setUp(self)
        self.args['distribution'] = adept_core.DIST_OPAL

    def test_no_aoi(self):
        try:
            del self.args['area_of_influence_uri']
        except KeyError:
            # Key not in args dict anyways
            pass
        adept_core.execute(self.args)

class SmokeTest_ES(SmokeTest):
    def setUp(self):
        natcap.opal.i18n.language.set('es')
        SmokeTest.setUp(self)



import logging
import json
import os

from osgeo import ogr
import shapely
import shapely.ops
import shapely.speedups
import shapely.wkb
import shapely.prepared
import pygeoprocessing

import preprocessing
import offsets
import utils

LOGGER = logging.getLogger('natcap.opal.analysis')

#    # now, aggregate over the correct polygons.
#    LOGGER.debug('Aggregating impacts')
#    for service_name, impact_raster in services:
#        field_name = "%s_sum" % service_name
#        aggregate_stats(impact_raster, out_vector_uri, 'FID', field_name)

def aggregate_stats(raster_uri, aggregate_vector, id_field, target_field,
        percent_to_stream=None):
    # aggregate stats over impact raster
    # loop through the correct component of the stats object and set the value
    # of the correct field.

    utils.assert_files_exist([raster_uri, aggregate_vector])
    target_field = target_field[:8]

    # collect stats under the aggregate polygon
    raster_stats = pygeoprocessing.aggregate_raster_values_uri(raster_uri,
        aggregate_vector, id_field).total

    if percent_to_stream is not None:
        LOGGER.debug('Adjusting aggregates by percent_to_stream %s',
            percent_to_stream)
        pts_per_impacts = pygeoprocessing.aggregate_raster_values_uri(
            percent_to_stream, aggregate_vector, id_field).pixel_max
        LOGGER.debug('pts_per_impacts: %s', pts_per_impacts)

        #print pts_per_impacts
        #print raster_stats

        LOGGER.debug('adjusting by percent-to-stream')
        adjusted_stats = {}
        for fid, pixel_sum in raster_stats.iteritems():
            try:
                percent_contribution = pts_per_impacts[fid]
                if percent_contribution is None:
                    percent_contribution = 0.0
            except KeyError:
                print 'KeyError!!!'
                percent_contribution = 0.0

            adjusted_stats[fid] = pixel_sum * percent_contribution
        LOGGER.debug('Finished adjusting by percent-to-stream')
    else:
        # If we're not adjusting by percent-to-stream, just use existing
        # raster_stats dictionary instead.
        adjusted_stats = raster_stats


    LOGGER.debug('Opening aggregate vector for editing')
    # I'm assuming that a side effect is ok in this case.
    out_vector = ogr.Open(aggregate_vector, 1)
    out_layer = out_vector.GetLayer()

    # create a new field in the output vector for the target field.
    LOGGER.debug('Adding new field "%s"', target_field)
    new_field_defn = ogr.FieldDefn(target_field, ogr.OFTReal)
    out_layer.CreateField(new_field_defn)

    # map FID field value to polygon index
    fid_fields = {}
    num_features = out_layer.GetFeatureCount()
    for feature_id in xrange(num_features):
        feature = out_layer.GetFeature(feature_id)
        fid_field_value = feature.GetField('FID')
        fid_fields[fid_field_value] = feature_id

    # I only care about the sum under the polygons ('total' attribute)
    LOGGER.debug('Number of fields analyzed: %s', len(raster_stats))
    for field_id, pixel_sum in adjusted_stats.iteritems():
        # Assuming that the field ID is the feature ID.
        feature = out_layer.GetFeature(fid_fields[field_id])
        feature.SetField(target_field, pixel_sum)
        out_layer.SetFeature(feature)

    out_layer.SyncToDisk()
    out_layer = None
    ogr.DataSource.__swig_destroy__(out_vector)
    out_vector = None

    return adjusted_stats

def calculate_biodiversity_impact(permitting_area_ds_uri, ecosystems_ds_uri):
    """Generates a dictionary indexed by ecosystem name that has both the
        impacted area and required offset area for mitigation.

        permitting_area_ds_uri -

        ecosystems_ds_uri -

        returns a dictionary of the format
            {
                ecosystem_name_1: {
                    'area': impacted area (float),
                    'mitigation_area'; required_mitigation area (float)
                    }
                ecosystem_name_2...
            }
        """


    utils.assert_files_exist([permitting_area_ds_uri, ecosystems_ds_uri])

    #1) Load the permitting_area dataset into a shapely permitting_area_polygon
    permitting_area_ds = ogr.Open(permitting_area_ds_uri)
    permitting_area_layer = permitting_area_ds.GetLayer()
    polygon_list = []
    for feature_index in xrange(permitting_area_layer.GetFeatureCount()):
        feature = permitting_area_layer.GetFeature(feature_index)
        geometry = feature.GetGeometryRef()
        polygon_list.append(shapely.wkb.loads(geometry.ExportToWkb()))
    permitting_area_polygon = shapely.ops.cascaded_union(polygon_list)

    if not permitting_area_polygon.is_valid:
        raise IOError('Permitting area is not a valid polygon')


    #2) Make a spatial index?
    if shapely.speedups.available:
        LOGGER.debug('Enabling shapely speedups')
        shapely.speedups.enable()

    impacts_schema = [f.GetName() for f in permitting_area_layer.schema]

    #3) Loop through each feature in permitting_area ds and build a polygon out of it
    biodiversity_impacts = {}
    ecosystems_ds = ogr.Open(ecosystems_ds_uri)
    ecosystems_ds_layer = ecosystems_ds.GetLayer()

    # determine if we want to also record ES impacts.
    ecosystems_schema = set(map(lambda f: f.GetName(),
        ecosystems_ds_layer.schema))
    _services_map = {
        'mitrat_c': 'carbon',
        'mitrat_s': 'sediment',
        'mitrat_n': 'nitrogen',
        'mitrat_cus': 'custom'
    }
    services = []
    for column_name, service_name in _services_map.iteritems():
        if column_name in ecosystems_schema:
            services.append(service_name)
    services = sorted(services)

    for feature_index in xrange(ecosystems_ds_layer.GetFeatureCount()):
        feature = ecosystems_ds_layer.GetFeature(feature_index)
        _get = lambda x: feature.GetField(x)
        ecosystem_type = _get('ecosystem')
        impact_factor = _get('mit_ratio')
        try:
            parcel_lci = _get('LCI')
        except ValueError:
            parcel_lci = None

        try:
            threat_score = _get('Threat')
        except ValueError:
            threat_score = None

        try:
            richness_score = _get('Richness')
        except ValueError:
            richness_score = None

        geometry = feature.GetGeometryRef()
        polygon = shapely.wkb.loads(geometry.ExportToWkb())
        prepared_polygon = shapely.prepared.prep(polygon)
        if prepared_polygon.intersects(permitting_area_polygon):
            LOGGER.debug('Permitting area intersects.  Calculating area overlap')
            intersection = permitting_area_polygon.intersection(polygon)

            # dividing by 10,000 gives us ha.
            impacted_area = intersection.area / 10000.0
            mitigation_area = impacted_area * impact_factor
            impacted_parcel_area = (polygon.area / 10000.0)
            try:
                # If we have already found a parcel with this ecosystem type,
                # then we want to update the tracked amounts of biodiversity
                # that has been impacted and is required for offsets.
                ecosystem_impacts = biodiversity_impacts[ecosystem_type]
                ecosystem_impacts['impacted_area'] += impacted_area
                ecosystem_impacts['mitigation_area'] += mitigation_area

                previous_min = ecosystem_impacts['min_impacted_parcel_area']
                if previous_min > impacted_parcel_area:
                    ecosystem_impacts['min_impacted_parcel_area'] = impacted_parcel_area

                ecosystem_impacts['min_lci'] = max(
                    ecosystem_impacts['min_lci'], parcel_lci)

                ecosystem_impacts['max_threat'] = max(
                    ecosystem_impacts['max_threat'], threat_score)

                ecosystem_impacts['min_richness'] = min(
                    ecosystem_impacts['min_richness'], richness_score)

                ecosystem_impacts['patches_impacted'] += 1

            except KeyError:
                # If we have not yet impacted a parcel with this ecosystem
                # type, then create a new dictionary for tracking its relevant
                # information.
                ecosystem_impacts = {
                    'impacted_area': impacted_area,
                    'mitigation_ratio': impact_factor,
                    'mitigation_area': mitigation_area,
                    'min_impacted_parcel_area': impacted_parcel_area,
                    'min_lci': parcel_lci,
                    'max_threat': threat_score,
                    'min_richness': richness_score,
                    'patches_impacted': 1,
                }

            # Use the mean mitigation ratio.
            impacted_area = ecosystem_impacts['impacted_area']
            if impacted_area == 0:
                impacted_area = 1.0

            ecosystem_impacts['mitigation_ratio'] = (
                float(ecosystem_impacts['mitigation_area']) / float(impacted_area))

            biodiversity_impacts[ecosystem_type] = ecosystem_impacts

            LOGGER.debug('Overlaps: %s',
                biodiversity_impacts[ecosystem_type]['impacted_area'])
            LOGGER.debug('Required offset: %s Ha',
                biodiversity_impacts[ecosystem_type]['mitigation_area'])
    LOGGER.debug('Done.')
    return biodiversity_impacts

# TODO: make temp_file optional
def percent_overlap(offset_sites, municipalities, temp_file, pop_col):
    utils.assert_files_exist([offset_sites, municipalities])

    # first, get a set of municipalities polygons that intersect
    preprocessing.locate_intersecting_polygons(municipalities, offset_sites,
        temp_file)

    offset_vector = ogr.Open(offset_sites)
    offset_layer = offset_vector.GetLayer()
    offset_schema = [f.GetName() for f in offset_layer.schema]

    # These are columns in the vector that we want to skip, probably because we
    # want the offset_data dictionary key to be something custom.
    known_columns = ['ecosystem']

    all_offsets = {}
    for offset_feature in offset_layer:
        offset_fid = offset_feature.GetFID()
        offset_geometry = offset_feature.GetGeometryRef()
        offset_area = offset_geometry.Area()
        shapely_polygon = offsets.build_shapely_polygon(offset_feature)
        offset_data = {
#            'sediment': offset_feature.GetField('sediment'),
#            'nutrient': offset_feature.GetField('nutrient'),
#            'carbon': offset_feature.GetField('carbon'),
#            'distance': offset_feature.GetField('distance'),
            'area': offset_area,
            'polygon': shapely_polygon,
            'prep_polygon': shapely.prepared.prep(shapely_polygon),
        }
        if 'ecosystem' in offset_schema:
            offset_data['ecosystem'] = offset_feature.GetField('ecosystem')

        for field_name, field_value in offset_feature.items().iteritems():
            if field_name not in known_columns:
                offset_data[field_name] = field_value

        all_offsets[offset_fid] = offset_data

    temp_vector = ogr.Open(temp_file)
    temp_layer = temp_vector.GetLayer()


    final_data = {}
    for offset_fid, offset_data in all_offsets.iteritems():
        offset_area = offset_data['polygon'].area
        final_offset_data = {
#            'Sediment': offset_data['sediment'],
#            'Nitrogen': offset_data['nutrient'],
#            'Carbon': offset_data['carbon'],
#            'distance': offset_data['distance'],
#            'ecosystem': offset_data['ecosystem'],
#            'area': offset_data['area'],
            'municipalities': {}
        }

        for offset_key, offset_value in offset_data.iteritems():
            if offset_key in ['polygon', 'prep_polygon']:
                continue

            if offset_key.lower() == 'nutrient':
                offset_key = 'nitrogen'

            final_offset_data[offset_key.capitalize()] = offset_value

        for municipality in temp_layer:
            m_polygon = offsets.build_shapely_polygon(municipality)
            m_name = municipality.GetField(pop_col)

            # check to see if there's overlap.
            if offset_data['prep_polygon'].intersects(m_polygon):
                inter_area = offset_data['polygon'].intersection(m_polygon).area

                overlap_ratio = inter_area / offset_area

                if overlap_ratio > 0:
                    final_offset_data['municipalities'][m_name] = overlap_ratio
        temp_layer.ResetReading()
        final_data[offset_fid] = final_offset_data
    return final_data

def vectors_intersect(vector_1_uri, vector_2_uri):
    """Take in two OGR vectors (we're assuming that they're in the same
    projection) and test to see if their geometries intersect.  Return True of
    so, False if not.

    vector_1_uri - a URI to an OGR vector
    vector_2_uri - a URI to an OGR vector

    Returns True or False"""

    utils.assert_files_exist([vector_1_uri, vector_2_uri])

    LOGGER.debug('Opening vector %s', vector_1_uri)
    basename_1 = os.path.basename(vector_1_uri)
    vector_1 = ogr.Open(vector_1_uri)
    layer_1 = vector_1.GetLayer()

    LOGGER.debug('Opening vector %s', vector_2_uri)
    basename_2 = os.path.basename(vector_2_uri)
    vector_2 = ogr.Open(vector_2_uri)
    layer_2 = vector_2.GetLayer()

    for feature_1 in layer_1:
        prep_polygon = offsets.build_shapely_polygon(feature_1, prep=True)

        for feature_2 in layer_2:
            polygon = offsets.build_shapely_polygon(feature_2)
            if prep_polygon.intersects(polygon):
                fid_1 = feature_1.GetFID()
                fid_2 = feature_2.GetFID()
                LOGGER.debug('%s (fid %s) and %s (fid %s) intersect',
                    basename_1, fid_1, basename_2, fid_2)
                return True
        layer_2.ResetReading()
    LOGGER.debug('No Features intersect.')
    return False

def features_in_vector(vector_uri):
    """Get the number of features in this vector.

    Args:
        vector_uri (string): The string filepath of the OGR vector to query

    Returns:
        The integer number of features in the vector."""

    vector = ogr.Open(vector_uri)
    layer = vector.GetLayer()
    num_features = int(layer.GetFeatureCount())
    return num_features





import unittest
import json
import tempfile
import os

import shapely
import shapely.geometry
import numpy
from osgeo import gdal

import adept.tests

class TestUtilsVectorTest(unittest.TestCase):
    def test_vector(self):
        polygon_a = shapely.geometry.Polygon([(1,1), (4,1), (4,4), (1,4)])

        fields = {
            'field_int': int,
            'field_str': str,
            'field_float': float
        }
        field_values = {
            'field_int': 1,
            'field_str': 'hello',
            'field_float': 3.14159
        }

        # writes to a geojson file for easy parsing.
        uri = adept.tests.vector([polygon_a], adept.tests.COLOMBIA_SRS, fields,
            [field_values])

        json_vector = json.load(open(uri))
        self.assertEqual(len(json_vector['features']), 1)
        self.assertEqual(json_vector['features'][0]['properties'], field_values)

        adept.tests.cleanup(uri)

    def test_raster(self):
        matrix = numpy.array([
            [1.0, 1.0, 1.0],
            [2.0, 2.0, 2.0],
            [3.0, 3.0, 3.0],
        ], dtype=numpy.float32)
        raster_uri = adept.tests.raster(matrix, adept.tests.COLOMBIA_SRS,
                adept.tests.COLOMBIA_GEOTRANSFORM(30, -30), 1.0)

        ds = gdal.Open(raster_uri)
        ds_band = ds.GetRasterBand(1)
        ds_type = ds_band.DataType
        self.assertEqual(ds_type, gdal.GDT_Float32)

        # get the output matrix, verify it's equal to the matrix we input.
        out_array = ds.ReadAsArray()
        numpy.testing.assert_array_equal(matrix, out_array)

        adept.tests.cleanup(raster_uri)

        # Verify that the raster can be written to a known path.
        temp_dir = tempfile.mkdtemp()
        out_file = os.path.join(temp_dir, 'new_file.tif')
        adept.tests.raster(matrix, adept.tests.COLOMBIA_SRS,
            adept.tests.COLOMBIA_GEOTRANSFORM(30, -30), 1.0, filename=out_file)

        self.assertTrue(os.path.exists(out_file))
        ds = gdal.Open(out_file)
        ds_band = ds.GetRasterBand(1)
        ds_type = ds_band.DataType
        self.assertEqual(ds_type, gdal.GDT_Float32)
        numpy.testing.assert_array_equal(matrix, ds_band.ReadAsArray())
        adept.tests.cleanup(temp_dir)


    def test_colombia_geotransform(self):
        expected_gt = [444720, 30, 0, 3751320, 0, -30]
        self.assertEqual(adept.tests.COLOMBIA_GEOTRANSFORM(30, -30),
            expected_gt)




import os
import json
import hashlib
import logging
from types import UnicodeType
from types import DictType
import multiprocessing
import shutil
import random
import tempfile
import sys
import distutils.sysconfig
import zipfile

from osgeo import gdal
from osgeo import ogr
from natcap.invest.sdr import sdr
from natcap.invest.nutrient import nutrient
from natcap.invest.carbon import carbon_combined as carbon
import pygeoprocessing
import numpy
import scipy

import preprocessing
import utils

LOGGER = logging.getLogger('natcap.opal.static_maps')
NODATA = 999999.0

COLOMBIA_BARE_LUCODE = 301
COLOMBIA_PAVED_LUCODE = 89

MODELS = {
    'carbon': {
        'module': carbon,
        'landcover_key': 'lulc_cur_uri',
        'target_raster': os.path.join('output', 'tot_C_cur.tif'),
        'watersheds_key': None,
    },
    'sediment': {
        'module': sdr,
        'landcover_key': 'lulc_uri',
        'target_raster': os.path.join('output', 'sed_export.tif'),
        'watersheds_key': 'watersheds_uri',
    },
    'nutrient': {
        'module': nutrient,
        'landcover_key': 'lulc_uri',
        'target_raster': os.path.join('output', 'n_export.tif'),
        'watersheds_key': 'watersheds_uri',
    }
}


def _write_future_json(workspace, future_type):
    """Write the future type to a json object at workspace/future_type.json.

        workspace - a URI to the static maps workspace.
        future_type - a string indicating the future type.  One of 'protection'
            or 'restoration'

    Returns nothing."""

    json_uri = os.path.join(workspace, 'future_type.json')
    json.dumps({'future_type': future_type}, open(json_uri, 'w'), indent=4,
               sort_keys=True)


def execute(args):
    """Entry point for generating static sediment maps.

        args - a python dictionary with the following attributes:
            workspace_dir (required)
            landuse_uri (required)
            landcover_code (required)
            future_type (required) - either 'protection' or 'restoration'
            model_name (required) - either 'carbon' or 'sediment'
            do_parallelism (optional) - Boolean.  Assumed to be False.
            fut_landuse_uri (optional) - URI.  If not present, a future
                landcover scenario will not be calculated.

            If model_name is 'sediment', these keys may be provided.  If they
            are not provided, default values will be assumed.
            dem_uri (optional)
            erosivity_uri (optional)
            erodibility_uri(optional)
            watersheds_uri (optional)
            biophysical_table_uri (optional)
            threshold_flow_accumulation (optional)
            slope_threshold (optional)
            sediment_threshold_table_uri (optional)

            If model_name is either 'sediment' or 'nutrient', the following
            is optional:
            num_simulations - an int indicating the number of impact sites
                that should be simulated per watershed.  If this key is not
                provided, static map quality estimates will be skipped.
            """

    for key in ['workspace_dir', 'landuse_uri', 'paved_landcover_code',
                'bare_landcover_code', 'model_name']:
        assert key in args, "Args is missing a key: %s" % key

    assert args['model_name'] in ['carbon', 'sediment', 'nutrient'], (
        'Model name must be one of "carbon", "sediment", or "nutrient",'
        'not %s' % args['model_name'])

    if not os.path.exists(args['workspace_dir']):
        os.makedirs(args['workspace_dir'])
        LOGGER.debug('Creating new workspace: %s', args['workspace_dir'])

    # create a logging handler and write its contents out to the logfile.
    log_handler = logging.FileHandler(os.path.join(args['workspace_dir'],
                                                   'logfile.txt'))
    log_formatter = logging.Formatter(
        fmt=(
            '%(asctime)s %(name)-18s '
            '%(threadName)-10s %(levelname)-8s %(message)s'),
        datefmt='%m/%d/%Y %H:%M:%S ')
    log_handler.setFormatter(log_formatter)
    LOGGER.addHandler(log_handler)

    temp_dir = os.path.join(args['workspace_dir'], 'tmp')
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
    tempfile.tempdir = temp_dir

    if args['model_name'] == 'carbon':
        args['lulc_cur_uri'] = args['landuse_uri']
    elif args['model_name'] in ['nutrient', 'sediment']:
        args['lulc_uri'] = args['landuse_uri']

    model_args = {}
    default_config = get_static_data_json(args['model_name'])
    for key, value in default_config.iteritems():
        if key in args:
            model_args[key] = args[key]
        else:
            model_args[key] = value

    if args['model_name'] == 'sediment':
        # we're ok with the landuse dictionary key.
        # clip the DEM to the landcover raster
        new_dem_uri = os.path.join(
            model_args['workspace_dir'],
            'clipped_dem.tif')
        LOGGER.debug('Current DEM: %s', model_args['dem_uri'])
        LOGGER.debug('Saving clipped DEM to %s', new_dem_uri)
        _clip_dem(model_args['dem_uri'], model_args['lulc_uri'], new_dem_uri)
        model_args['dem_uri'] = new_dem_uri
    elif args['model_name'] == 'nutrient':
        # filter the watersheds to just those that intersect with the LULC.
        new_watersheds_uri = os.path.join(model_args['workspace_dir'],
                                          'watersheds_filtered.shp')
        preprocessing.filter_by_raster(
            model_args['lulc_uri'],
            model_args['watersheds_uri'],
            new_watersheds_uri)
        model_args['watersheds_uri'] = new_watersheds_uri
        model_args['soil_depth_uri'] = model_args[
            'depth_to_root_rest_layer_uri']
        try:
            model_args['eto_uri'] = args['potential_evapotranspiration']
        except KeyError:
            # MAFE uses the internal eto_uri, so we can skip.
            LOGGER.debug('No key "potential_evapotranspiration"')

    LOGGER.debug(
        'Using these model args:\n%s',
        json.dumps(
            model_args,
            sort_keys=True,
            indent=4))

    # now, run the sediment model on the input landcover
    LOGGER.info('Running the model on the original landscape')
    original_workspace = os.path.join(args['workspace_dir'],
                                      '%s_base' % args['model_name'])
    if not os.path.exists(original_workspace):
        os.makedirs(original_workspace)
        LOGGER.debug('Making workspace for base scenario: %s',
                     original_workspace)

    LOGGER.debug('Original workspace: %s', original_workspace)
    execute_model(args['model_name'], args['landuse_uri'], original_workspace,
                  model_args)
    base_raster = os.path.join(original_workspace,
                               MODELS[args['model_name']]['target_raster'])

    try:
        do_parallelism = args['do_parallelism']
        LOGGER.debug('Process-based parallelism enabled')
    except KeyError:
        LOGGER.debug('Process-based parallelism disabled')
        do_parallelism = False

    try:
        num_simulations = int(args['num_simulations'])
        LOGGER.debug('User requested to do %s simulations per watershed',
                     num_simulations)
    except KeyError:
        num_simulations = None
        LOGGER.debug('Skipping impact simulations')

    processes = []
    for impact_type in ['paved', 'bare']:
        continue  # Only want to do the future scenario right now.
        LOGGER.debug('Starting calculations for impact %s', impact_type)
        impact_code = args['%s_landcover_code' % impact_type]

        impact_workspace = os.path.join(args['workspace_dir'], impact_type)
        static_map_uri = os.path.join(
            args['workspace_dir'], '%s_%s_static_map.tif' %
            (args['model_name'], impact_type))

        # Carbon is the only known ES that has inverted values.
        invert = True if args['model_name'] == 'carbon' else False

        if do_parallelism:
            process = multiprocessing.Process(
                target=build_static_map,
                args=(
                    args['model_name'],
                    args['landuse_uri'],
                    impact_code,
                    static_map_uri,
                    base_raster,
                    model_args,
                    impact_workspace),
                kwargs={
                    'num_simulations': num_simulations,
                    'invert': invert})
            processes.append(process)
            process.start()
        else:
            build_static_map(
                args['model_name'],
                args['landuse_uri'],
                impact_code,
                static_map_uri,
                base_raster,
                model_args,
                impact_workspace,
                num_simulations=num_simulations,
                invert=invert)

    # Build the static protection map if the user has provided a future
    # landcover scenario.
    LOGGER.info('Found a future landcover.  Building (%s) map.',
                args['future_type'])
    if 'fut_landuse_uri' not in args:
        LOGGER.debug('No custom future lulc found.  Clipping default.')
        # if the user has not provided a custom lulc, we should clip the
        # existing future lulc to the size of the user-defined current lulc.
        common_data = get_common_data_json()
        future_landuse_uri = os.path.join(
            args['workspace_dir'],
            'future_landuse.tif')
        _clip_dem(common_data['future_landcover'], args['landuse_uri'],
                  future_landuse_uri)
    else:
        future_landuse_uri = args['fut_landuse_uri']
    LOGGER.debug('Future landcover %s', future_landuse_uri)

    # determine whether the future, converted landcover should be inverted,
    # based on protection or restoration.
    _write_future_json(args['workspace_dir'], args['future_type'])
    if args['future_type'] == 'protection':
        invert = True
        future_tif_name = 'protect'
    else:
        invert = False
        future_tif_name = 'restore'

    # if we're building carbon static maps, the invert flag is opposite of what
    # other models would do, for all cases.
    if args['model_name'] == 'carbon':
        invert = not invert

    future_map_uri = os.path.join(
        args['workspace_dir'], '%s_%s_static_map.tif' %
        (args['model_name'], future_tif_name))
    future_workspace = os.path.join(args['workspace_dir'], future_tif_name)
    build_static_map(args['model_name'], args['landuse_uri'], future_landuse_uri,
                     future_map_uri, base_raster, model_args, future_workspace,
                     convert_landcover=False,  # just use the future landcover
                     num_simulations=num_simulations, invert=invert)
    LOGGER.info('Completed the future (%s) static map.', args['future_type'])

    # If we aren't doing parallelism, then this list has no elements in it.
    for process in processes:
        process.join()

    # If we just ran the nutrient model, we need to copy the appropriate
    # percent-to-stream rasters to the root static maps directory.
    # Do this copy by recompressing the GeoTiff using DEFLATE instead of LZW.
    # See issue 2910 (code.google.com/p/invest-natcap/issues/detail?id=2910)

    if args['model_name'] in ['nutrient']:
        fmt_string = os.path.join(args['workspace_dir'], '%s',
            '%s_converted' % args['model_name'], 'intermediate')
        pts_name = 'n_percent_to_stream.tif'
        fmt_string = os.path.join(fmt_string, pts_name)

        percent_to_streams = [
            (fmt_string % 'paved', '%s_paved_pts.tif' % args['model_name']),
            (fmt_string % 'bare', '%s_bare_pts.tif' % args['model_name']),
            (fmt_string % future_tif_name, '%s_%s_pts.tif' %
                (args['model_name'], future_tif_name)),
        ]
        for source_uri, dest_uri in percent_to_streams:
            dest_uri = os.path.join(args['workspace_dir'], dest_uri)
            LOGGER.debug('Copying %s to %s', source_uri, dest_uri)
            preprocessing.recompress_gtiff(source_uri, dest_uri, 'DEFLATE')

    LOGGER.debug('Completed creating the %s static maps', args['model_name'])


def raster_math(args):
    """Perform all of the raster math to create static maps from the input
    model run rasters.

    args - a python dictionary with the following attributes:
        workspace_dir - a URI to the output workspace
        name - a string name to be used in the filename of all output static maps
        base_uri - a URI to a GDAL raster of the base scenario's service values.
        paved_uri - a URI to a GDAL raster of the paved scenario's service values.
        bare_uri - a URI to a GDAL raster of the bare scenario's service values.
        future_uri - a URI to a GDAL raster of the future scenario's
            service values.
        future_type - a string, either 'protection' or 'restoration'

    Returns None, but writes the following files to disk:
        workspace/<name>_bare_static_map.tif
        workspace/<name>_paved_static_map.tif
        workspace/<name>_future_static_map.tif"""
    workspace = args['workspace_dir']
    name = args['name']
    base_uri = args['base_uri']
    paved_uri = args['paved_uri']
    bare_uri = args['bare_uri']
    future_uri = args['future_uri']
    future_scenario = args['future_type']
    _write_future_json(workspace, future_scenario)

    if not os.path.exists(workspace):
        LOGGER.debug('Creating output workspace %s', workspace)
        os.makedirs(workspace)

    def _ws_path(static_map_type):
        return os.path.join(workspace,
                            '%s_%s_static_map.tif' % (name, static_map_type))

    future_tif_base = 'protect' if future_scenario == 'protection' else 'restore'

    bare_sm_uri = _ws_path('bare')
    paved_sm_uri = _ws_path('paved')
    future_sm_uri = _ws_path(future_tif_base)

    # create the bare static map
    LOGGER.debug('Creating the bare static map %s', bare_sm_uri)
    subtract_rasters(bare_uri, base_uri, bare_sm_uri)

    # create the paved static map
    LOGGER.debug('Creating the paved static map %s', paved_sm_uri)
    subtract_rasters(paved_uri, base_uri, paved_sm_uri)

    # create the future static map
    LOGGER.debug(
        'Creating the %s static map %s',
        future_scenario,
        future_sm_uri)
    if future_scenario == 'protection':
        subtract_rasters(base_uri, future_uri, future_sm_uri)
    else:  # when args['future_type'] is 'restoration'
        subtract_rasters(future_uri, base_uri, future_sm_uri)

    LOGGER.debug('Finished creating the static maps')


def _clip_dem(dem_uri, lulc_uri, out_dem_uri):
    """Clip the input DEM to the LULC and save the resulting raster to the
    out_dem_uri."""
    utils.assert_files_exist([dem_uri, lulc_uri])
    nodata = pygeoprocessing.get_nodata_from_uri(dem_uri)
    pixel_size = pygeoprocessing.get_cell_size_from_uri(dem_uri)
    datatype = pygeoprocessing.get_datatype_from_uri(dem_uri)

    pygeoprocessing.vectorize_datasets([dem_uri, lulc_uri], lambda x, y: x,
        dataset_out_uri=out_dem_uri, datatype_out=datatype, nodata_out=nodata,
        pixel_size_out=pixel_size, bounding_box_mode='intersection',
        vectorize_op=False)

def convert_lulc(lulc_uri, new_code, out_uri):
    """Use vectorize_datasets to convert a land cover raster to a new landcover
    code and save it to an output dataset.

        lulc_uri - a uri to a GDAL landcover raster on disk.
        new_code - an integer landcover code
        out_uri - a uri to which the converted landcover will be written

        Returns nothing."""

    utils.assert_files_exist([lulc_uri])
    nodata = pygeoprocessing.get_nodata_from_uri(lulc_uri)

    # unless there's a nodata value, we want to change all pixels to the new
    # lulc code.
    def _convert(pixels):
        return numpy.where(pixels != nodata, new_code, nodata)

    pixel_size = pygeoprocessing.get_cell_size_from_uri(lulc_uri)
    datatype = pygeoprocessing.get_datatype_from_uri(lulc_uri)
    pygeoprocessing.vectorize_datasets([lulc_uri], _convert,
        dataset_out_uri=out_uri, datatype_out=datatype,
        nodata_out=nodata, pixel_size_out=pixel_size,
        bounding_box_mode='intersection', vectorize_op=False)

def unzip_static_zipfile(zipfile_uri):
    """Unzip the given file to the static maps folder."""

    utils.assert_files_exist([zipfile_uri])
    with zipfile.ZipFile('test.zip', 'r') as zip_archive:
        static_data_dir = os.path.join(os.getcwd(), 'data',
                                       'colombia_static_data')
        zip_archive.extractall(static_data_dir)


def execute_model(model_name, landcover_uri, workspace_uri, config=None):
    assert model_name in MODELS.keys()

    # if config is None, then we load the static_map parameters from the
    # correct internal json file.
    # if it's not None, then it must be a configuration dictionary.
    if config is None:
        config = get_static_data_json(model_name)
    else:
        assert isinstance(config, DictType), ("Found %s: %s" % (
            type(config), config))

    # - loop through each key in the configuration file.
    #    * if it's a path, make it relative to the CWD
    #    * if it's not a path, leave it alone.
    new_config = {}
    for key, value in config.iteritems():
        if isinstance(value, UnicodeType):
            try:
                new_value = float(value)
            except ValueError:
                if len(value) > 0:
                    if not os.path.exists(value):
                        new_value = os.path.join(os.getcwd(), '..', value)
                    else:
                        # if the user-defined file does exist, just use that.
                        new_value = value
                else:
                    new_value = value
        else:
            new_value = value
        new_config[key] = new_value

    # - add the new keys needed for the model run:
    #    * workspace_uri - use the URI passed in to this function.
    #    * landcover_uri - use the URI passed in to this function.
    new_config['workspace_dir'] = workspace_uri
    new_config[MODELS[model_name]['landcover_key']] = landcover_uri

    LOGGER.debug(
        'Executing model with arguments: %s',
        json.dumps(
            new_config,
            sort_keys=True,
            indent=4))
    MODELS[model_name]['module'].execute(new_config)


def build_static_map(
        model_name,
        landcover_uri,
        landcover_code,
        static_map_uri,
        base_run,
        config=None,
        workspace=None,
        convert_landcover=True,
        num_simulations=None,
        invert=False):
    """Build the static map for the target ecosystem service.  Currently assumes
    we're doing sediment only.

        model_name - a string.  Must be a key in MODELS.
        landcover_uri - a URI to the user's input landcover.  A raster.
        landcover_core - an integer landcover code to convert to.
        static_map_uri - a URI to where the output static map should be written
        base_run - a URI to the output map from a base run of the target model.
        config=None - a python dictionary with arguments for the target model.  If
            None, the defaults will be loaded from internal defaults.
        workspace=None - the output workspace to which the model runs should be
            written.  If None, they will be saved to a temporary folder and
            deleted at the end of the static map generation.
        base_run=None - A uri to a workspace of the target model.
        num_simulations=None - number of simulations to run per watershed.  If
            None, no simulations will be run.
        invert=False - whether to invert the subtraction.  When invert==False,
            the static map produced will be the difference of `base_run` -
            `converted`, where `converted` is the converted or input landcover.
            When invert==True, the static map produced will be the differece
            of `converted` - `base_run`.
    """
    assert invert in [True, False], '%s found instead' % type(invert)
    assert model_name in MODELS.keys()
    LOGGER.info('Building static map for the %s model', model_name)

    if workspace is not None:
        LOGGER.debug('Using workspace %s', workspace)
        if not os.path.exists(workspace):
            LOGGER.debug('Creating workspace folder %s', workspace)
            os.makedirs(workspace)
    else:
        workspace = pygeoprocessing.temporary_folder()
        LOGGER.debug('Writing model workspace data to %s', workspace)

    # convert the LULC to the correct landcover code
    if convert_landcover:
        converted_lulc = os.path.join(workspace, 'converted_lulc.tif')
        LOGGER.info('Creating converted landcover raster: %s', converted_lulc)
        convert_lulc(landcover_uri, landcover_code, converted_lulc)
        landcover_label = str(landcover_code)
    else:
        converted_lulc = landcover_uri
        landcover_label = 'transformed'

    LOGGER.info('Running the model on the converted landcover')
    # run the sediment model on the converted LULC
    target_raster = MODELS[model_name]['target_raster']
    converted_workspace = os.path.join(workspace, '%s_converted' % model_name)
    LOGGER.debug('Converted workspace: %s', converted_workspace)
    execute_model(model_name, converted_lulc, converted_workspace, config)
    converted_es_map = os.path.join(converted_workspace, target_raster)

    # subtract the two rasters.
    # If we're running the carbon model, the service we're measuring (carbon
    # storage) has positive values being 'good'  For sediment and nutrient,
    # however, positive values represent sediment/nutrient export to the steam,
    # which is 'bad'.  For sediment and nutrient, we want to invert the result.
    # This is done by just reversing the order of the subtraction.
    LOGGER.info('Subtracting the two rasters.  Invert=%s', invert)
    if invert is True:
        subtract_rasters(converted_es_map, base_run, static_map_uri)
    else:
        subtract_rasters(base_run, converted_es_map, static_map_uri)

    if num_simulations is not None:
        if workspace is None:
            # if the user created the workspace in a temporary folder, create a
            # folder in CWD for the quality workspace.
            workspace = os.getcwd()

        if config is None:
            # in case the user has not provided the config dictionary
            config = get_static_data_json(model_name)

        watersheds = config[MODELS[model_name]['watersheds_key']]

        simulation_workspace = os.path.join(workspace, 'simulations_%s' %
                                            landcover_label)
        test_static_map_quality(
            base_run,
            static_map_uri,
            landcover_uri,
            landcover_code,
            watersheds,
            model_name,
            simulation_workspace,
            config,
            num_simulations,
            invert=invert)

        simulations_csv = os.path.join(simulation_workspace,
                                       'impact_site_simulation.csv')
        out_png = os.path.join(simulation_workspace, 'simulations.png')
        graph_it(simulations_csv, out_png)
    LOGGER.info('Finished')


def subtract_rasters(raster_a, raster_b, out_uri):
    utils.assert_files_exist([raster_a, raster_b])
    LOGGER.debug('Subtracting rasters %s and %s', raster_a, raster_b)
    LOGGER.debug('Saving difference to %s', out_uri)
    pixel_size = pygeoprocessing.get_cell_size_from_uri(raster_a)
    nodata = pygeoprocessing.get_nodata_from_uri(raster_b)
    datatype = pygeoprocessing.get_datatype_from_uri(raster_b)
    LOGGER.debug('Output pixel size: %s', pixel_size)
    LOGGER.debug('Output nodata value: %s', nodata)
    LOGGER.debug('Output datatype: %s', datatype)

    pygeoprocessing.vectorize_datasets([raster_a, raster_b],
        lambda c, o: numpy.where(c != nodata, numpy.subtract(c, o), nodata),
        dataset_out_uri = out_uri,
        datatype_out=datatype, nodata_out=nodata,
        pixel_size_out=pixel_size, bounding_box_mode='intersection',
        vectorize_op=False)


def get_static_data_json(model_name):
    """Get the absolute path to the static data JSON file for the target model.

        model_name - a python string model name.  Must be a key in MODELS.

        Returns a python dictionary with the configuration."""
    assert model_name in MODELS.keys()

    if model_name == 'sediment':
        model_name = 'sdr'

    json_name = '%s_parameters.json' % model_name
    return _load_json(json_name)


def get_common_data_json(data_dir=None):
    """Return a dictionary with paths to common data (e.g. hydrozones, etc.).

        Returns a python dictionary"""
    common_data_name = 'common_data.json'
    config = _load_json(common_data_name)

    if data_dir is None:
        if getattr(sys, 'frozen', False):
            data_dir = os.path.dirname(sys.executable)
        else:
            data_dir = os.getcwd()

    def _render_dict(dictionary):
        output_dict = {}
        for key, item in dictionary.iteritems():
            if isinstance(item, DictType):
                rendered_item = _render_dict(item)
            else:
                rendered_item = os.path.join(data_dir, item)
            output_dict[key] = rendered_item
        return output_dict

    return _render_dict(config)


def _load_json(filename):
    """Fetch a json file from the adept package's static_data folder.

        Returns a python dictionary of the json data found in the target json
        file."""
    if getattr(sys, 'frozen', False):
        # we are running in a |PyInstaller| bundle
        # basedir = sys._MEIPASS  #suggested by pyinstaller
        basedir = os.path.join(sys._MEIPASS, 'natcap', 'opal')
    else:
        basedir = os.path.dirname(__file__)

    LOGGER.debug('__file__ == %s', __file__)
    LOGGER.debug('sys.executable == %s', sys.executable)
    LOGGER.debug('site-packages == %s', distutils.sysconfig.get_python_lib())
    LOGGER.debug('Looking for common static data in %s', basedir)
    config_file = os.path.join(basedir, 'static_data', filename)
    config = json.load(open(config_file))
    return config


def get_json_md5(json_uri):
    """Get an md5 hash for the python dictionary stored in a json file.  This
    function tries to be aware of whether a value points to a file on disk.
    When this happens, we fetch an MD5sum for the file and use that in the
    digest instead of the URI.

    Returns a python string with an MD5sum digest of the json object."""

    utils.assert_files_exist(json_uri)
    LOGGER.debug('Loading json from %s', json_uri)

    config = json.load(open(json_uri))
    config_md5sum = hashlib.md5()

    # assume that this is a flat dictionary.
    for key, value in config.iteritems():
        # if the value is a unicode string that is a URI to a file on disk, we
        # want to get the md5sum of the file and use that as the value in the
        # config's md5sum.  If it's not a URI to a file on disk, we'll just use
        # the value as is.
        if isinstance(value, UnicodeType):
            if os.path.exists(value):
                LOGGER.debug('Value %s is a URI', value)
                file_handler = open(value, 'rb')
                file_md5 = hashlib.md5()
                for chunk in iter(lambda: file_handler.read(2**20), ''):
                    file_md5.update(chunk)
                value = file_md5.hexdigest()

        LOGGER.debug('Updating digest with %s: %s', key, value)
        config_md5sum.update(key)
        config_md5sum.update(value)

    return config_md5sum.hexdigest()


def clip_raster_to_watershed(in_raster, ws_vector, out_uri, clip_raster=None):
    """Clip the input raster to ws_vector, saving the output raster to out_uri.
        in_raster - a URI to an input GDAL raster.
        ws_vector - a URI to an OGR vector that contains a single polygon of a
            watershed.
        out_uri - a URI to where the output raster should be saved.
    """
    datatype = pygeoprocessing.get_datatype_from_uri(in_raster)
    nodata = pygeoprocessing.get_nodata_from_uri(in_raster)
    pixel_size = pygeoprocessing.get_cell_size_from_uri(in_raster)

    if clip_raster is not None:
        rasters = [in_raster, clip_raster]
        clip_nodata = pygeoprocessing.get_nodata_from_uri(clip_raster)

        def operation(in_values, clip_values):
            return numpy.where(
                clip_values == clip_nodata,
                clip_nodata,
                in_values)
    else:
        rasters = [in_raster]
        operation = lambda x: x

    pygeoprocessing.vectorize_datasets(
        rasters,
        operation,
        out_uri,
        datatype,
        nodata,
        pixel_size,
        'intersection',
        dataset_to_align_index=0,
        aoi_uri=ws_vector,
        vectorize_op=False)


def make_random_impact_vector(new_vector, base_vector, side_length):
    """Create a new vector with a single, squarish polygon.  This polygon will
    be created within the spatial envelope of the first polygon in base_vector.
    The new squarish polygon will have a side length of side_length.

        new_vector - a URI to the new vector to be created. The new vector will
            be an ESRI Shapefile.
        base_vector - a URI to the vector we'll use as a base (for its spatial
            information).
        side_length - a python int or float describing the side length of the
            new polygon to be created.

        Returns nothing."""
    base_datasource = ogr.Open(base_vector)
    base_layer = base_datasource.GetLayer()
    base_feature = base_layer.GetFeature(0)
    base_geometry = base_feature.GetGeometryRef()
    spat_ref = base_layer.GetSpatialRef()

    # feature_extent = [xmin, xmax, ymin, ymax]
    feature_extent = base_geometry.GetEnvelope()
    print feature_extent

    driver = ogr.GetDriverByName('ESRI Shapefile')
    datasource = driver.CreateDataSource(new_vector)
    uri_basename = os.path.basename(new_vector)
    layer_name = str(os.path.splitext(uri_basename)[0])
    layer = datasource.CreateLayer(layer_name, spat_ref, ogr.wkbPolygon)

    # Add a single ID field
    field = ogr.FieldDefn('id', ogr.OFTInteger)
    layer.CreateField(field)

    while True:
        poly_ring = ogr.Geometry(type=ogr.wkbLinearRing)
        bbox_width = feature_extent[1]-feature_extent[0]
        bbox_height = feature_extent[3]-feature_extent[2]

        rand_width_percent = random.random()
        xmin = feature_extent[0] + bbox_width * rand_width_percent
        xmax = xmin + side_length

        # Make it squarish
        rand_height_percent = random.random()
        ymin = feature_extent[2] + bbox_height * rand_height_percent
        ymax = ymin + side_length

        print feature_extent
        print xmin, xmax, ymin, ymax

        poly_ring.AddPoint(xmin, ymin)
        poly_ring.AddPoint(xmin, ymax)
        poly_ring.AddPoint(xmax, ymax)
        poly_ring.AddPoint(xmax, ymin)
        poly_ring.AddPoint(xmin, ymin)
        polygon = ogr.Geometry(ogr.wkbPolygon)
        polygon.AddGeometry(poly_ring)

        # See if the watershed contains the permitting polygon
        contained = base_geometry.Contains(polygon)
        print contained
        if contained:
            break

    feature = ogr.Feature(layer.GetLayerDefn())
    feature.SetGeometry(polygon)
    feature.SetField(0, 1)
    layer.CreateFeature(feature)

    feature = None
    layer = None


def get_watershed_id(watershed_uri):
    """Get the ID from the watershed specified by the user.

        watershed_uri (string) - A String URI to the watershed vector.  This
            vector is assumed to have exactly one watershed polygon in it.

        Returns an int watershed ID."""

    # get this watershed's ws_id
    watershed_vector = ogr.Open(watershed_uri)
    watershed_layer = watershed_vector.GetLayer()
    watershed = watershed_layer.GetFeature(0)
    watershed_id = watershed.GetField('ws_id')
    LOGGER.debug('This watershed\'s ws_id: %s', watershed_id)
    return watershed_id


def test_static_map_quality(
        base_run,
        base_static_map,
        landuse_uri,
        impact_lucode,
        watersheds_uri,
        model_name,
        workspace,
        config,
        num_iterations=5,
        clean_workspaces=False,
        start_ws=0,
        start_impact=0,
        invert=None):
    """Test the quality of the provided static map.

    Args:
        base_run (filepath): The base run of the target model on the base lulc.
        base_static_map (filepath): The static map generated from the difference
            between the base_run raster and the entire landscape converted to the
            target impact type.
        landuse_uri (filepath): A URI to the LULC used for the base static map.
        impact_lucode (int or float): The numeric land use code to use to convert
            the underlying lulc raster to the target impact type.
        watersheds_uri (filepath): A filepath to the watersheds vector to use for
            testing.  Must have a column of integers in its attribute table
            labeled "ws_id".
        model_name (string): The string model name to run.
        workspace (filepath): The path to the folder to use as a workspace.  If
            this folder does not already exist, it will be created.
        config (dict): The arguments dictionary to use for running the model.
            See `static_maps.execute_model()` for details, or else
            `natcap/opal/static_data/<model_name>_parameters.json` for sample
            argument dictionaries (albeit serialized as JSON).
        num_iterations=5 (int, optional): The number of simulated impacts to run per
            watershed.
        clean_workspace=False (boolean, optional): Whether to remove the
            workspace before starting to test the inputs.
        start_ws=0 (int, optional): The watershed index to start on.  If 0, all
            watersheds will be tested.  Useful for resuming testing after
            failure (such as when running out of disk space).
        start_impact=0 (int, optional): The integer impact ID to start on.
            This must be less than `num_interations`.
        invert=None (boolean): Whether to invert the static map calculation.

    Returns:
        Nothing.
    """
    assert invert in [True, False], '%s found instead' % type(invert)
    old_tempdir = tempfile.tempdir
    temp_dir = os.path.join(workspace, 'tmp')  # for ALL tempfiles
    tempfile.tempdir = temp_dir  # all tempfiles will be saved here.

    # make a copy of the configuration dictionary so that we don't modify it
    # accidentally.
    config = config.copy()

    # make all the folders we know about at the moment
    pygeoprocessing.create_directories([workspace, temp_dir])

    # Open a logfile so we can incrementally write model data we care about
    logfile_uri = os.path.join(workspace, 'impact_site_simulation.csv')
    logfile = open(logfile_uri, 'a')
    labels = ['ws_id', 'Impact ID', 'Impact Area', 'Static Estimate',
              'InVEST Estimate', 'Estimate Ratio']
    logfile.write("%s\n" % ','.join(labels))
    logfile.close()

    lulc_nodata = pygeoprocessing.get_nodata_from_uri(landuse_uri)
    lulc_pixel_size = pygeoprocessing.get_cell_size_from_uri(landuse_uri)

    # limit the watersheds to just those that intersect the input lulc.
    current_watersheds = os.path.join(temp_dir, 'current_watersheds.shp')
    preprocessing.filter_by_raster(landuse_uri, watersheds_uri,
        current_watersheds, clip=True)

    # get the sediment export from the base raster, passed in from the user.
    # calculate for each watershed, so I can access these later.
    #base_export = pygeoprocessing.aggregate_raster_values_uri(
    #    base_run, current_watersheds, 'ws_id', 'sum').total
    #LOGGER.debug('All watershed ids: %s', base_export.keys())

    # split the watersheds so I can use each watershed as an AOI for the
    # correct model later on.
    watersheds_dir = os.path.join(workspace, 'watershed_vectors')
    split_watersheds = split_datasource(
        current_watersheds,
        watersheds_dir,
        ['ws_id'])

    for ws_index, watershed_uri in enumerate(split_watersheds):
        if ws_index < start_ws:
            LOGGER.debug(
                'Watershed %s is less than start index %s. skipping',
                ws_index, start_ws)
            continue

        watershed_workspace = os.path.join(
            workspace, 'watershed_%s' % ws_index)
        if not os.path.exists(watershed_workspace):
            os.makedirs(watershed_workspace)

        # get this watershed's ws_id
        watershed_vector = ogr.Open(watershed_uri)
        watershed_layer = watershed_vector.GetLayer()
        watershed = watershed_layer.GetFeature(0)
        watershed_id = watershed.GetField('ws_id')
        LOGGER.debug('This watershed\'s ws_id: %s', watershed_id)

        watershed_lulc = os.path.join(watershed_workspace,
            'watershed_lulc.tif')
        lulc_datatype = pygeoprocessing.get_datatype_from_uri(landuse_uri)
        pygeoprocessing.vectorize_datasets([landuse_uri], lambda x: x,
            watershed_lulc, lulc_datatype, lulc_nodata, lulc_pixel_size,
            'intersection', dataset_to_align_index=0, aoi_uri=watershed_uri,
            vectorize_op=False)

        ws_base_export_uri = os.path.join(watershed_workspace,
            'watershed_' + os.path.basename(base_run))
        base_nodata = pygeoprocessing.get_nodata_from_uri(base_run)
        base_pixel_size = pygeoprocessing.get_cell_size_from_uri(base_run)
        base_export_datatype = pygeoprocessing.get_datatype_from_uri(base_run)
        pygeoprocessing.vectorize_datasets([base_run], lambda x: x,
            ws_base_export_uri, base_export_datatype, base_nodata, base_pixel_size,
            'intersection', dataset_to_align_index=0, aoi_uri=watershed_uri,
            vectorize_op=False)
        base_ws_export = pygeoprocessing.aggregate_raster_values_uri(
            ws_base_export_uri, watershed_uri, 'ws_id',
            'sum').total[watershed_id]

        # if the model uses watersheds, we only want to run the model using
        # the one current watershed.
        watersheds_key = MODELS[model_name]['watersheds_key']
        if watersheds_key is not None:
            config[watersheds_key] = watershed_uri

        watershed_lulc = os.path.join(watershed_workspace,
                                      'watershed_lulc.tif')
        clip_raster_to_watershed(landuse_uri, watershed_uri, watershed_lulc)
        watershed_base_workspace = os.path.join(watershed_workspace, 'base')
        execute_model(model_name, watershed_lulc, watershed_base_workspace, config)

        ws_base_export_uri = os.path.join(watershed_base_workspace,
                                          MODELS[model_name]['target_raster'])
        ws_base_static_map = os.path.join(
            watershed_workspace, 'watershed_' + os.path.basename(base_static_map))
        clip_raster_to_watershed(base_static_map, watershed_uri, ws_base_static_map)

        # If we're not in the starting watershed, then reset the starting index
        # of the impact site.
        start_impact = 0 if ws_index != start_impact else start_impact

        for run_number in range(start_impact, num_iterations):
            impact_site_length = random.uniform(500, 3000)
            impact_workspace = os.path.join(watershed_workspace,
                                            'random_impact_%s' % run_number)

            if os.path.exists(impact_workspace):
                shutil.rmtree(impact_workspace)
            os.makedirs(impact_workspace)

            # make a random impact vector somewhere in the current watershed.
            impact_site = os.path.join(
                impact_workspace,
                'impact_%s.shp' %
                run_number)
            make_random_impact_vector(impact_site, watershed_uri,
                                      impact_site_length)

            # convert the area under the impact to the correct landcover
            # code(s), run the target model and analyze the outputs.
            converted_landcover = os.path.join(impact_workspace,
                                               'converted_lulc.tif')
            # If the landcover is a string, we convert to the area under the
            # impact.  If the landcover is a number, that's the conversion
            # type.
            convert_impact(impact_site, watershed_lulc, impact_lucode,
                           converted_landcover, impact_workspace)
            execute_model(model_name, converted_landcover, impact_workspace,
                          config)
            estimates = aggregate_test_results(
                impact_workspace,
                model_name,
                watershed_uri,
                impact_site,
                ws_base_static_map,
                ws_base_export_uri,
                invert=invert)

            # ability to sort based on area of impact site.
            # also record which watershed this run is in, impact site ID as well
            impact_site_area = get_polygon_area(impact_site)
            values_to_write = [
                watershed_id,
                run_number,
                impact_site_area,
                estimates['static_est'],
                estimates['invest_est'],
                estimates['export_ratio'],
            ]
            logfile = open(logfile_uri, 'a')
            logfile.write("%s\n" % ','.join(map(str, values_to_write)))
            logfile.close()

def compute_impact_stats(impact_dir, model_name, watershed_vector,
                         base_ws_export, base_static_map):
    """Take an impact directory and the target model name and extract the
    correct information from it.

    impact_dir - a URI to a folder that has been used as an impact workspace.
    model_name - the string name of the model we're using.
    watershed_vector - a URI to an OGR vector of the watershed this impact
        belongs to.
    base_ws_export - the base watershed export (a number)
    base_static_map - a URI to the static map generated from the difference
        between the base sediment model run and when the landscape is converted
        completely over to the target impact type.

    Returns a python dictionary containing extracted stats about the impact."""

    impact_vector = 'impact_%s.shp' % os.path.basename(
        impact_dir).split('_')[-1]
    impact_site = os.path.join(impact_dir, impact_vector)
    print impact_site
    impact_site_area = get_polygon_area(impact_site)

    export_raster = os.path.join(impact_dir,
                                 MODELS[model_name]['target_raster'])

    # aggregate this impact over the target watershed.
    impact_ws_export = pygeoprocessing.aggregate_raster_values_uri(export_raster,
        watershed_vector, 'ws_id').total.values()[0]  # just get the only ws

    # get the sediment export from the base static map under the impacted area.
    # only 1 feature in the impacted area, so we access this number with index
    # 1.
    static_estimate = pygeoprocessing.aggregate_raster_values_uri(
        base_static_map, impact_site, 'id').total[1]

    # If we're running the nutrient model, multiply the sum of the
    # sed_export by the max percent_to_stream under the impact site.
    if model_name in ['nutrient']:
        LOGGER.info('Adjusting export by the % to stream')

        # the percent-to-stream raster for Nitrogen is named
        # "n_percent_to_stream.tif", sediment is just "percent_to_stream.tif"
        # nutrient percent-to-stream is prefixed by 'n_'
        pts_prefix = 'n_'

        percent_to_stream = os.path.join(impact_dir,
            'intermediate', '%spercent_to_stream.tif' % pts_prefix)
        max_percent = pygeoprocessing.aggregate_raster_values_uri(
            percent_to_stream, impact_site,
            'id').pixel_max[1]
        if max_percent is None:
            LOGGER.debug('Max percent is None, setting to 0')
            max_percent = 0.0
        static_estimate = static_estimate * max_percent
    else:
        LOGGER.info('Not running a routed model, running %s', model_name)

    invest_estimate = impact_ws_export - base_ws_export
    #invest_estimate = base_ws_export - impact_ws_export

    export_ratio = static_estimate / invest_estimate

    return {
        'impact_dir': impact_dir,
        'impact_site_area': impact_site_area,
        'static_estimate': static_estimate,
        'invest_estimate': invest_estimate,
        'export_ratio': export_ratio,
        'impact_ws_export': impact_ws_export,
    }


def get_polygon_area(vector):
    # ONLY returns the area of the first polygon.
    datasource = ogr.Open(vector)
    layer = datasource.GetLayer()
    feature = layer.GetFeature(0)
    geometry = feature.GetGeometryRef()
    area = geometry.Area()

    geometry = None
    feature = None
    layer = None
    datasource = None
    return area


def graph_it(log_file, out_file):
    import matplotlib
    matplotlib.use('Agg')  # for rendering plots without $DISPLAY set.
    import matplotlib.pyplot as plt

    LOGGER.info('Creating graph from results at %s', log_file)
    LOGGER.debug('Saving image to %s', out_file)
    all_rows = []
    out_of_bounds = []
    opened_log_file = open(log_file)
    opened_log_file.next()  # skip the column headers.
    for line in opened_log_file:
        try:
            values = map(float, line.split(','))
        except ValueError as error:
            # when there's a column with string data that can't be cast to
            # a float (like a column header), skip the row.
            LOGGER.warn(error)
            continue

        ws_id, run_num, impact_area, static_est, invest_est, ratio = values

#        if ratio > 3 or ratio < -3:
#            out_of_bounds.append(ratio)
#        else:
#            all_rows.append((impact_area, ratio))
        all_rows.append((impact_area, ratio))

    # smoother with 95 % confidence intervals
    all_rows = sorted(all_rows, key=lambda x: x[0])
    areas = [r[0] for r in all_rows]
    ratios = [r[1] for r in all_rows]

#    LOGGER.debug('These values were outliers: %s', out_of_bounds)
    plt.plot(areas, ratios, 'ro')
    plt.xlabel('Impact Site Area (m^2)')
    plt.ylabel('(Static Est. / InVEST Est)')

    areas_np = numpy.array(areas)
    ratios_np = numpy.array(ratios)

    n = len(ratios_np)
    t = scipy.linspace(0, max(areas), n)

    # Linear regressison -polyfit - polyfit can be used other orders polys
    (ar, br) = scipy.polyfit(areas_np, ratios_np, 1)
    xr = scipy.polyval([ar, br], t)

    plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    plt.plot(t, xr, 'g--')  # plot the linear regression line.
    plt.savefig(out_file)


def split_datasource(ds_uri, workspace, include_fields=[],
                     template_str='feature_%s.shp'):
    """Split the input OGR datasource into a list of datasources, each with a
    single layer containing a single feature.
        ds_uri - a URI to an OGR datasource.
        workspace - a folder to which the output vectors should be saved.
        include_fields=[] - a list of string fields to be copied from the source
            datsource to the destination datasources.
        template_str - a template string with a placeholder for a single value.
            All file uris will be named according to this pattern.
    Returns a list of URIs, one for each new vector created."""

    if os.path.exists(workspace):
        shutil.rmtree(workspace)

    os.makedirs(workspace)

    LOGGER.debug('Opening vector at %s', ds_uri)
    ds = ogr.Open(ds_uri)
    driver_string = 'ESRI Shapefile'

    LOGGER.debug('Splitting datasource into separate shapefiles')
    LOGGER.debug('Vectors will be saved to %s', workspace)
    output_vectors = []
    for layer in ds:
        layer_defn = layer.GetLayerDefn()
        for feature in layer:
            print 'starting new feature'
            uri_index = feature.GetFID()
            new_vector_uri = os.path.join(workspace, template_str %
                                          uri_index)
            output_vectors.append(new_vector_uri)

            LOGGER.debug('Creating new shapefile at %s' % new_vector_uri)
            ogr_driver = ogr.GetDriverByName(driver_string)
            temp_shapefile = ogr_driver.CreateDataSource(new_vector_uri)
            LOGGER.debug('SRS: %s, %s', layer.GetSpatialRef(),
                         type(layer.GetSpatialRef()))

            layer_name = os.path.splitext(os.path.basename(
                new_vector_uri))[0]
            if isinstance(layer_name, UnicodeType):
                LOGGER.debug('Decoding layer name %s to ASCII', layer_name)
                #layer_name = layer_name.decode('utf-8')
                layer_name = str(layer_name)

            LOGGER.debug('Layer name: %s', layer_name)

            temp_layer = temp_shapefile.CreateLayer(
                layer_name, layer.GetSpatialRef())
            temp_layer_defn = temp_layer.GetLayerDefn()

            for field_index in range(layer_defn.GetFieldCount()):
                original_field = layer_defn.GetFieldDefn(field_index)
                output_field = ogr.FieldDefn(original_field.GetName(),
                                             original_field.GetType())
                temp_layer.CreateField(output_field)

            # Create the obligatory ID field.
            # If I don't create the ID field, I can't properly select other
            # fields later on, when I need to set their values.
            id_field = ogr.FieldDefn('id', ogr.OFTInteger)
            temp_layer.CreateField(id_field)

            # Create the new feature with all of the characteristics of the old
            # field except for the fields.  Those are brought along separately.
            LOGGER.debug('Creating new feature with duplicate geometry')
            feature_geom = feature.GetGeometryRef()
            temp_feature = ogr.Feature(temp_layer_defn)
            temp_feature.SetFrom(feature)

            # Since there's only one feature in this shapefile, set id to 0.
            id_field_index = temp_feature.GetFieldIndex('id')
            temp_feature.SetField(id_field_index, 0)

            LOGGER.debug('Copying over fields %s', include_fields)
            for field_index in range(layer_defn.GetFieldCount()):
                field_defn = layer_defn.GetFieldDefn(field_index)
                field = field_defn.GetName()
                LOGGER.debug('Adding field "%s"', field)

                # Create the new field in the temp feature
                field_type = field_defn.GetType()
                LOGGER.debug('Field type=%s', field_type)

                LOGGER.debug('Copying field "%s" value to new feature',
                             field)
                temp_feature.SetField(field, feature.GetField(field))

            temp_layer.CreateFeature(temp_feature)
            temp_layer.SyncToDisk()

            temp_layer = None
            temp_shapefile = None

    layer.ResetReading()
    layer = None
    ds = None
    ogr_driver = None

    LOGGER.debug('Finished creating the new shapefiles')
    return output_vectors


def convert_impact(impact_uri, base_lulc, impacted_value, converted_lulc_uri,
                   workspace):
    """Convert the area under the impact vector to be the value of
        impact_value.

        impact_uri (string) - a filepath to an impact site vector on disk.
        base_lulc (string) - a filepath to the base lulc on disk.
        impacted_value (string or int) - The value to convert to.  If an int,
            the value under the impact site will be this landcover code.  If
            a string, the value under the impact site will be the pixel values
            of this raster under the impact site.
        converted_lulc_uri (string) - a filepath to where the converted raster
            should be stored.
        workspace (string) - a filepath to a folder where some output rasters
            will be written.

        Returns nothing."""

    # Create a raster mask for the randomized impact site.
    # Any non-nodata pixels underneath the impact site are marked by 1.
    impact_mask = os.path.join(workspace, 'impact_mask.tif')
    lulc_nodata = pygeoprocessing.get_nodata_from_uri(base_lulc)
    lulc_pixel_size = pygeoprocessing.get_cell_size_from_uri(base_lulc)
    lulc_datatype = pygeoprocessing.get_datatype_from_uri(base_lulc)

    def mask_op(values):
        return numpy.where(values != lulc_nodata, 1.0,
                           lulc_nodata)

    pygeoprocessing.vectorize_datasets(
        [base_lulc],
        mask_op,
        impact_mask,
        lulc_datatype,
        lulc_nodata,
        lulc_pixel_size,
        'intersection',
        dataset_to_align_index=0,
        aoi_uri=impact_uri,
        vectorize_op=False)


    if isinstance(impacted_value, basestring):
        LOGGER.debug('Converting values to those of %s', impacted_value)
        def _convert_impact(mask_values, lulc_values, impacted_lulc_values):
            """Convert values under the mask to the future lulc values."""
            return numpy.where(mask_values == 1, impacted_lulc_values,
                               lulc_values)
        rasters_list = [impact_mask, base_lulc, impacted_value]
    else:
        LOGGER.debug('Converting values to scalar: %s', impacted_value)
        def _convert_impact(mask_values, lulc_values):
            """Convert values under the mask to the scalar impacted value."""
            return numpy.where(mask_values == 1, impacted_value,
                               lulc_values)
        rasters_list = [impact_mask, base_lulc]

    pygeoprocessing.vectorize_datasets(
        rasters_list, _convert_impact, converted_lulc_uri,
        lulc_datatype, lulc_nodata, lulc_pixel_size, 'union',
        dataset_to_align_index=0, vectorize_op=False)


def aggregate_test_results(impact_workspace, model_name, watershed_uri,
                           impact_site, base_static_map, base_export, invert):
    # get the target raster for the selected ecosystem service.
    export = os.path.join(impact_workspace,
                          MODELS[model_name]['target_raster'])

    def _mask_out_pixels(in_raster, comp_raster, out_raster):
        comp_nodata = pygeoprocessing.get_nodata_from_uri(comp_raster)
        pixel_size = pygeoprocessing.get_cell_size_from_uri(comp_raster)

        def _pixel_mask(_in_values, _out_values):
            return numpy.where(_in_values == comp_nodata,
                comp_nodata, _out_values)

        pygeoprocessing.vectorize_datasets([comp_raster, in_raster],
            _pixel_mask, out_raster, gdal.GDT_Float32, comp_nodata,
            pixel_size, 'union', dataset_to_align_index=0,
            vectorize_op=False)

    # mutually mask out the impacted/base export rasters.
    masked_impact_export = os.path.join(impact_workspace,
                                        'masked_impacted_export.tif')
    masked_base_export = os.path.join(impact_workspace,
                                      'masked_base_export.tif')
    _mask_out_pixels(export, base_export, masked_impact_export)
    _mask_out_pixels(base_export, export, masked_base_export)

    # Aggregate the sediment export from this impact simulation over
    # the target watershed
    impact_ws_export = pygeoprocessing.aggregate_raster_values_uri(
        masked_impact_export, watershed_uri, 'ws_id').total.values()[0]

    # Get the sediment export from the static map under the impacted area.
    # only 1 feature in the impactd area, so we access that number with
    # index 1.
    static_estimate = pygeoprocessing.aggregate_raster_values_uri(
        base_static_map, impact_site, 'id').total[1]

    # Get the watershed's base export from the masked version of the
    # watershed's export raster.
    watershed_id = get_watershed_id(watershed_uri)
    base_ws_export = pygeoprocessing.aggregate_raster_values_uri(
        masked_base_export, watershed_uri, 'ws_id').total[watershed_id]

    LOGGER.warning('NOT adjusting by %%-to-stream. model=%s', model_name)
    # This conditional makes the outputs all
    # represent the same thing: positive values are desireable,
    # negative values are not desireable.
    if invert:
        invest_estimate = impact_ws_export - base_ws_export
    else:
        invest_estimate = base_ws_export - impact_ws_export


    export_ratio = static_estimate / invest_estimate

    return {
        'static_est': static_estimate,
        'invest_est': invest_estimate,
        'export_ratio': export_ratio,
        'base_export': base_ws_export,
        'impacted_export': impact_ws_export,
    }


def clip_static_map(map_uri, aoi_uri, out_uri):
    """Clip the input static map by the single polygon in aoi_uri.  Saves the
    output raster to out_uri.  Values outside of the aoi will be set to nodata.

    map_uri - a URI to a GDAL raster.
    aoi_uri - a URI to an OGR vector.  May only contain one polygon.
    out_uri - the URI to which the output raster should be saved.

    Returns nothing."""

    nodata = pygeoprocessing.get_nodata_from_uri(map_uri)
    pixel_size = pygeoprocessing.get_cell_size_from_uri(map_uri)

    pygeoprocessing.vectorize_datasets([map_uri], lambda x: x, out_uri,
        gdal.GDT_Float32, nodata, pixel_size, 'intersection', aoi_uri=aoi_uri,
        vectorize_op=False)
